# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utilities for commons nodes.
"""
from numbers import Number
from typing import Union

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_argument_iterable,
    check_argument_operator,
    check_argument_orthogonal_projection_operator,
    check_argument_real_vector,
    check_numeric_numpy_array,
    check_sample_times,
)
from scipy.sparse import spmatrix

from .node_data import (
    Pwc,
    Stf,
    Tensor,
)

TensorLike = Union[np.ndarray, Tensor]
TensorLikeOrFunction = Union[np.ndarray, Tensor, Pwc, Stf]
NumericOrFunction = Union[float, complex, np.ndarray, Tensor, Pwc, Stf]
_IntType = (int, np.integer)


def get_broadcasted_shape(*shapes):
    """
    Return the shape resulting of broadcasting multiple shapes,
    or None if they're not broadcastable.

    The shapes are broadcastable if, for each dimension starting from the end,
    they all have either the same size or a size 1.

    Parameters
    ----------
    *shapes : tuple[int]
        Shapes of the objects.

    Returns
    -------
    tuple[int] or None
        The resulting broadcasted shape if the shapes are broadcastable, otherwise None.
    """

    # Obtain largest length of shapes.
    max_shape_len = max(len(shape) for shape in shapes)

    # Prepend ones to shorter shapes.
    extended_shapes = [(1,) * (max_shape_len - len(shape)) + shape for shape in shapes]

    # Check that the shapes are broadcastable:
    # for each position, all shapes hold either the same value or a 1 and another value.
    broadcasted_shape = []
    for dimensions in zip(*extended_shapes):
        dim_set = set(dimensions)
        if len(dim_set) > 2 or (len(dim_set) == 2 and 1 not in dim_set):
            return None
        # The product of each valid set will be either 1 or the non-1 value.
        broadcasted_shape.append(int(np.prod(list(dim_set))))

    return tuple(broadcasted_shape)


def validate_broadcasted_shape(x_shape, y_shape, x_name, y_name):
    """
    Get the resulting broadcasted shape for two input shapes,
    throwing an error if they're not broadcastable.

    Parameters
    ----------
    x_shape : tuple[int]
        One of the shapes to be broadcasted.
    y_shape : tuple[int]
        The other of the shapes to be broadcasted.
    x_name : str
        The name of the variable whose shape is `x_shape`, used for the error
        message in case the shapes aren't broadcastable.
    y_name : str
        The name of the variable whose shape is `y_shape`, used for the error
        message in case the shapes aren't broadcastable.

    Returns
    -------
    tuple[int]
        The shape of the broadcasted array.

    Raises
    ------
    QctrlArgumentsValueError
        If the two shapes aren't broadcastable.
    """

    shape = get_broadcasted_shape(x_shape, y_shape)
    check_argument(
        shape is not None,
        f"The shapes of {x_name} and {y_name} must be broadcastable.",
        {f"{x_name} shape": x_shape, f"{y_name} shape": y_shape},
    )
    return shape


def validate_function_output_shapes(
    x_batch_shape, x_value_shape, y_batch_shape, y_value_shape, validate_value_shape
):
    """
    Get the output batch and value shape for two input shapes of Pwcs/Stfs.
    The names of the variables are assumed to be x and y when reporting errors.

    Parameters
    ----------
    x_batch_shape : tuple[int]
        The batch shape of the first object.
    x_value_shape : tuple[int]
        The value shape of the first object.
    y_batch_shape : tuple[int]
        The batch shape of the second object.
    y_value_shape : tuple[int]
        The value shape of the second object.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple]
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.

    Returns
    -------
    tuple[int], tuple[int]
        The batch and value shapes of the output Pwc/Stf.

    Raises
    ------
    QctrlArgumentsValueError
        if the two objects aren't compatible.
    """
    batch_shape = validate_broadcasted_shape(
        x_batch_shape, y_batch_shape, "x (batch)", "y (batch)"
    )
    value_shape = validate_value_shape(x_value_shape, y_value_shape, "x", "y")

    return batch_shape, value_shape


def validate_tensor_and_function_output_shapes(
    t_shape,
    f_batch_shape,
    f_value_shape,
    t_name,
    f_name,
    validate_value_shape,
    tensor_first=True,
):
    """
    Get the output batch and value shape for an input tensor and an input Pwc/Stf.

    Parameters
    ----------
    t_shape : tuple[int]
        The shape of the tensor.
    f_batch_shape : tuple[int]
        The batch shape of the Pwc/Stf.
    f_value_shape : tuple[int]
        The value shape of the Pwc/Stf.
    t_name : str
        The name of the tensor variable, used for the error message in case the shapes aren't
        compatible.
    f_name : str
        The name of the function variable, used for the error message in case the shapes aren't
        compatible.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple]
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.
    tensor_first : bool, optional
        Whether the Tensor is the leftmost parameter. Defaults to True.

    Returns
    -------
    tuple[int], tuple[int]
        The batch and value shapes of the output Pwc/Stf.

    Raises
    ------
    QctrlArgumentsValueError
        if the two objects aren't compatible.
    """
    if tensor_first:
        value_shape = validate_value_shape(t_shape, f_value_shape, t_name, f_name)
    else:
        value_shape = validate_value_shape(f_value_shape, t_shape, f_name, t_name)

    return f_batch_shape, value_shape


def validate_shape(tensor_like, tensor_like_name):
    """
    Return the shape of a scalar, np.ndarray, scipy.sparse.spmatrix, or Tensor node.

    Parameters
    ----------
    tensor_like : number or np.ndarray or scipy.sparse.spmatrix or Tensor
        The object whose shape you want to obtain.
    tensor_like_name : str
        The name of the `tensor_like`, used for error message in case the
        input object is not valid.

    Returns
    -------
    tuple[int]
        The tuple with the size of each dimension of `tensor_like`.

    Raises
    ------
    QctrlArgumentsValueError
        if the input is neither a scalar, a NumPy array, nor a Tensor.
    """

    if hasattr(tensor_like, "shape"):
        return tuple(tensor_like.shape)

    if np.isscalar(tensor_like):
        return ()

    raise QctrlArgumentsValueError(
        f"The type of {tensor_like_name} is not valid.", {tensor_like_name: tensor_like}
    )


def validate_batch_and_value_shapes(tensor, tensor_name):
    """
    Return the batch and value shapes of Pwc or Stf.

    Parameters
    ----------
    tensor : Pwc or Stf
        The NodeData for the Pwc or Stf whose batch and value shapes
        you want to obtain.
    tensor_name : str
        The name of the Pwc or Stf, used for the error message in
        case `tensor` doesn't have a value shape.

    Returns
    -------
    tuple[tuple[int]]
        A tuple with a tuple that represents the batch shape, and a tuple
        that represents the value shape, in this sequence.

    Raises
    ------
    QctrlArgumentsValueError
        if the input is neither a scalar, a Pwc, nor an Stf
    """

    if hasattr(tensor, "value_shape") and hasattr(tensor, "batch_shape"):
        return tuple(tensor.batch_shape), tuple(tensor.value_shape)

    if hasattr(tensor, "value_shape"):
        return (), tuple(tensor.value_shape)

    raise QctrlArgumentsValueError(
        f"The type of {tensor_name}={tensor} must be Pwc or Stf.", {tensor_name: tensor}
    )


def validate_hamiltonian(hamiltonian, hamiltonian_name):
    """
    Check whether a Pwc, Stf or SparsePwc contains values that are Hamiltonians.

    Hamiltonians are two-dimensional and square.

    Parameters
    ----------
    hamiltonian : Pwc or Stf or SparsePwc
        The Hamiltonian to be tested.
    hamiltonian_name : str
        The name of the Hamiltonian, used in the error message.
    """
    value_shape = getattr(hamiltonian, "value_shape", ())

    check_argument(
        len(value_shape) == 2,
        "The shape of the Hamiltonian must have 2 dimensions.",
        {hamiltonian_name: hamiltonian},
        extras={f"{hamiltonian_name}.value_shape": value_shape},
    )
    check_argument(
        value_shape[-1] == value_shape[-2],
        "The dimensions of the Hamiltonian must have equal sizes.",
        {hamiltonian_name: hamiltonian},
        extras={f"{hamiltonian_name}.value_shape": value_shape},
    )


def validate_ms_shapes(ion_count, ld_values, ld_name, rd_values, rd_name):
    """
    Check if the shapes of the Mølmer–Sørensen parameters are correct.

    The correct shapes for the input parameters of Mølmer–Sørensen gate are
    ``(3, ion_count, ion_count)`` for the Lamb–Dicke parameters and
    ``(3, ion_count)`` for the relative detunings.

    Parameters
    ----------
    ion_count : int
        The number of ions in the chain.
    ld_values : np.ndarray
        The input values of the Lamb–Dicke parameters.
    ld_name : str
        The name of the argument that holds the Lamb–Dicke parameters.
    rd_values : np.ndarray
        The input values of the relative detunings.
    rd_name : str
        The name of the argument that holds the relative detunings.
    """
    check_numeric_numpy_array(ld_values, ld_name)
    check_numeric_numpy_array(rd_values, rd_name)

    ld_shape = validate_shape(ld_values, ld_name)
    rd_shape = validate_shape(rd_values, rd_name)

    check_argument(
        ld_shape == (3, ion_count, ion_count),
        "The Lamb–Dicke parameters must have shape (3, ion_count, ion_count).",
        {ld_name: ld_values},
        extras={"ion_count": ion_count, f"{ld_name}.shape": ld_shape},
    )
    check_argument(
        rd_shape == (3, ion_count),
        "The relative detunings must have shape (3, ion_count).",
        {rd_name: rd_values},
        extras={"ion_count": ion_count, f"{rd_name}.shape": rd_shape},
    )


def validate_batched_operator(operator, name):
    """
    Check if the input is a valid batched operator.

    Parameters
    ----------
    operator : Tensor or np.ndarray
        An operator.
    name : str
        Name of the operator.

    Returns
    -------
    tuple[int], tuple[int]
        The batch shape and value shape of the operator, if all checks are passed.

    Raises
    ------
    QctrlArgumentsValueError
        If operator doesn't have numeric element, or is not square.
    """
    check_numeric_numpy_array(operator, name)
    operator_shape = validate_shape(operator, name)
    check_argument(
        len(operator_shape) >= 2,
        f"The {name} must be at least two dimensional.",
        {name: operator},
    )
    check_argument(
        operator_shape[-1] == operator_shape[-2],
        f"The {name} must be a square in the last two dimensions.",
        {name: operator},
        extras={f"{name} shape": operator_shape},
    )
    return operator_shape[:-2], operator_shape[-2:]


def check_density_matrix_shape(density_matrix, name):
    """
    Check the shape of the input density matrix.

    Parameters
    ----------
    density_matrix : Tensor or np.ndarray
        A density matrix.
    name : str
        Name of the density matrix.
    """
    batch_shape, value_shape = validate_batched_operator(density_matrix, name)
    check_argument(
        len(batch_shape) in [0, 1],
        f"The {name} must be 2D or 3D with the first axis as the batch dimension.",
        {name: density_matrix},
        extras={"density matrix shape": batch_shape + value_shape},
    )


def check_oqs_hamiltonian(hamiltonian, system_dimension, system_name):
    """
    Check whether an open quantum system (OQS) Hamiltonian is valid.

    Parameters
    ----------
    hamiltonian : Pwc or SparsePwc
        Effective system Hamiltonian.
    system_dimension : int
        The dimension of the system.
    system_name : str
        The name of the system for error message.
    """
    check_argument(
        hamiltonian.value_shape == (system_dimension, system_dimension),
        f"The dimension of the Hamiltonian must be compatible with the dimension of {system_name}.",
        {"hamiltonian": hamiltonian},
        extras={
            "hamiltonian dimension": hamiltonian.value_shape,
            "system dimension": system_dimension,
        },
    )


def check_lindblad_terms(lindblad_terms, system_dimension, system_name):
    """
    Check whether Lindblad terms are valid, and whether the number of terms
    is greater than zero.

    Parameters
    ----------
    lindblad_terms : list[tuple[float, np.ndarray or spmatrix]]
        The list of Lindblad terms.
    system_dimension : int
        The dimension of the system.
    system_name : str
        The name of the system for error message.
    """
    check_argument_iterable(lindblad_terms, "lindblad_terms")
    check_argument(
        len(lindblad_terms) > 0,
        "You must provide at least one Lindblad term.",
        {"lindblad_terms": lindblad_terms},
    )

    for index, term in enumerate(lindblad_terms):
        check_argument(
            isinstance(term, tuple) and len(term) == 2,
            "Lindblad terms must be a list of tuples and each tuple must have decay rate and "
            "operator as the elements.",
            {"lindblad_terms": lindblad_terms},
        )
        rate, operator = term
        check_argument(
            rate > 0,
            "The decay rate must be positive.",
            {"lindblad_terms": lindblad_terms},
        )
        check_argument(
            isinstance(operator, (spmatrix, np.ndarray, Tensor)),
            "Lindblad operator must either be a NumPy array, scipy sparse matrix, or Tensor.",
            {"lindblad_terms": lindblad_terms},
        )
        # Extra check for array.
        check_numeric_numpy_array(operator, f"lindblad_terms[{index}][1]")
        _ = validate_shape(operator, f"lindblad_terms[{index}][1]")
        check_argument_operator(operator, f"lindblad_terms[{index}][1]")
        check_argument(
            operator.shape[0] == system_dimension,
            "The dimension of Lindblad operator must be compatible with the dimension "
            f"of {system_name}.",
            {"lindblad_terms": lindblad_terms},
            extras={
                "Lindblad operator shape": operator.shape,
                "system dimension": system_dimension,
            },
        )


def validate_inputs_real_fourier_signal(
    fixed_frequencies, optimizable_frequency_count, randomized_frequency_count
):
    """
    Check if the inputs of real_fourier_pwc/stf_signal function are valid.
    """
    check_argument(
        (fixed_frequencies is not None)
        + (optimizable_frequency_count is not None)
        + (randomized_frequency_count is not None)
        == 1,
        "Exactly one of `fixed_frequencies`, `optimizable_frequency_count` and "
        "`randomized_frequency_count` must be provided.",
        {
            "fixed_frequencies": fixed_frequencies,
            "optimizable_frequency_count": optimizable_frequency_count,
            "randomized_frequency_count": randomized_frequency_count,
        },
    )
    if optimizable_frequency_count is not None:
        check_argument_integer(
            optimizable_frequency_count, "optimizable_frequency_count"
        )
        check_argument(
            optimizable_frequency_count > 0,
            "The number of optimizable frequencies (if provided) must be greater than zero.",
            {"optimizable_frequency_count": optimizable_frequency_count},
        )
    if randomized_frequency_count is not None:
        check_argument_integer(randomized_frequency_count, "randomized_frequency_count")
        check_argument(
            randomized_frequency_count > 0,
            "The number of randomized frequencies (if provided) must be greater than zero.",
            {"randomized_frequency_count": randomized_frequency_count},
        )


def validate_fock_state_input(dimension, level, offset):
    """
    Validate the input for the Fock state node.
    If all input are valid, this function returns the shape of the Fock state.
    """
    from_subsystem = not isinstance(dimension, _IntType)

    if from_subsystem:
        # when from subsystems
        # dimension must be a list of integers
        check_argument(
            _validate_list(dimension, _IntType),
            "Dimension must either be an integer or a list of integers.",
            {"dimension": dimension},
        )
        _dimension_arr = np.asarray(dimension)

        # offset must be a list of non-negative integers of the size len(dimension)
        if offset is not None:
            check_argument(
                isinstance(offset, list)
                and _validate_list(offset, _IntType, len(dimension)),
                "When constructed from subsystems, a non-default offset must be a list of integers "
                "and have the same size as dimension.",
                {"dimension": dimension, "offset": offset},
            )

        _offset_arr = (
            np.zeros_like(_dimension_arr) if offset is None else np.asarray(offset)
        )
        check_argument(
            np.all(_offset_arr >= 0), "Offset must be non-negative.", {"offset": offset}
        )

        # level must be List[int] or List[List[int]] or int
        if isinstance(level, list):
            check_level_list(level, dimension)
            _level_arr = np.asarray(level)
        elif isinstance(level, _IntType):
            _level_arr = np.repeat(level, len(dimension))
        else:
            raise QctrlArgumentsValueError(
                "Level must be an integer, a list of integers, or a list of lists of integers.",
                {"level": level},
            )

        check_argument(
            np.all(_level_arr >= _offset_arr),
            "Level must not be smaller than offset.",
            {"level": level, "offset": offset},
        )
        check_argument(
            np.all(_level_arr < (_dimension_arr + _offset_arr)),
            "Level must be smaller than dimension+offset.",
            {"level": level, "dimension": dimension, "offset": offset},
        )

        return _level_arr.shape[:-1] + (int(np.prod(dimension)),)

    # when direct setup
    # only integer is allowed for dimension
    check_argument(
        isinstance(dimension, _IntType),
        "Dimension must either be an integer or a list of integers.",
        {"dimension": dimension},
    )
    # offset must be non-negative integers if set
    _offset = 0 if offset is None else offset
    check_argument(
        isinstance(_offset, _IntType),
        "Dimension and offset must be the same type. They should either be "
        "integers or a list of integers.",
        {"dimension": dimension, "offset": offset},
    )
    check_argument(_offset >= 0, "Offset must be non-negative.", {"offset": offset})

    # level must be an integer or list of integers that are in [offset, dim+offset)
    check_argument(
        isinstance(level, _IntType) or _validate_list(level, _IntType, None),
        "If the Fock state is not constructed from subsystems, "
        "level must either be an integer or a list of integers.",
        {"dimension": dimension, "level": level},
    )
    level_arr = np.asarray(level)
    check_argument(
        np.all(level_arr >= _offset),
        "Level must not be smaller than offset.",
        {"offset": _offset, "level": level},
    )
    check_argument(
        np.all(level_arr < dimension + _offset),
        "Level must be smaller than dimension + offset.",
        {"dimension": dimension, "offset": _offset, "level": level},
    )

    return level_arr.shape + (dimension,)


def validate_field_operator_input(parameter, lower_bound, name):
    """
    Check the input parameter for creation/annihilation/number operator.
    """
    check_argument(
        isinstance(parameter, _IntType) and parameter >= lower_bound,
        f"{name} must be an integer and not less than {lower_bound}.",
        {name: parameter},
    )


def validate_coherent_state_alpha(alpha):
    """
    Check whether alpha is valid for generating coherent state and
    return the batch shape if alpha is valid. Otherwise, raise an error.
    """

    if isinstance(alpha, list):
        check_argument(
            _validate_list(alpha, Number),
            "Alpha must be either a number, a list of them, or a 1D NumPy array.",
            {"alpha": alpha},
        )
        return (len(alpha),)
    if isinstance(alpha, np.ndarray):
        check_argument(
            alpha.ndim == 1,
            "Alpha must be either a number, a list of them, or a 1D NumPy array.",
            {"alpha": alpha},
        )
        check_numeric_numpy_array(alpha, "alpha")
        return (len(alpha),)
    if isinstance(alpha, Number):
        return ()
    raise QctrlArgumentsValueError(
        "Alpha must be either a number, a list of them, or a 1D NumPy array.",
        {"alpha": alpha},
    )


def check_level_list(level, dimension):
    """
    If level is a list, it must be a list of integers and len(level) == len(dimension).
    If it's a nested list, each element can only be a list of integers of the size len(dimension).
    """
    if isinstance(level[0], _IntType):
        check_argument(
            _validate_list(level, _IntType, len(dimension)),
            "Level must be a list of integers with the same length as dimension.",
            {"level": level, "dimension": dimension},
        )
    elif isinstance(level[0], list):
        for item in level:
            check_argument(
                isinstance(item, list)
                and _validate_list(item, _IntType, len(dimension)),
                "Each batch of level must be a list of integers with the same "
                "length as dimension.",
                {"level": level, "dimension": dimension},
            )
    else:
        raise QctrlArgumentsValueError(
            "Level must be an integer, a list of integers, or a list of lists of integers.",
            {"level": level},
        )


def _validate_list(list_, type_, length=None) -> bool:
    """
    Check list size and elements all have the same type.
    """
    if not isinstance(list_, list):
        return False
    if length is not None and len(list_) != length:
        return False
    for item in list_:
        if not isinstance(item, type_):
            return False
    return True


def check_operation_axis(axis, shape, tensor, tensor_name):
    """
    Certain Tensor operations are applied along the axis of the Tensor.
    The function checks
    1. whether the axis is consistent with the shape of Tensor.
    2. whether there is any repeated item in axis
    """

    if axis is None:
        return list(range(len(shape)))

    if isinstance(axis, _IntType):
        axis = [axis]
    else:
        check_argument(
            isinstance(axis, (list, tuple)),
            "Non-default axis must be an integer, or a list (tuple) of integers.",
            {"axis": axis},
        )
        for item in axis:
            check_argument(
                isinstance(item, _IntType),
                "Elements of axis must be integers.",
                {"axis": axis},
            )
        axis = list(axis)

    for i, dimension in enumerate(axis):
        check_argument(
            -len(shape) <= dimension < len(shape),
            f"Elements of axis must be valid axes of {tensor_name} (between {-len(shape)} "
            f"and {len(shape)-1}, inclusive).",
            {tensor_name: tensor, "axis": axis},
        )
        if dimension < 0:
            axis[i] = dimension + len(shape)

    check_argument(
        len(set(axis)) == len(axis),
        f"Elements of axis must refer to unique dimensions of {tensor_name}.",
        {tensor_name: tensor, "axis": axis},
    )

    return axis


def get_keepdims_operation_shape(shape, axis, keepdims):
    """
    Return the shape of the operations that can keep the dimension of the input tensor.
    """
    output_shape = []
    for i, size in enumerate(shape):
        if i not in axis:
            output_shape.append(size)
        elif keepdims:
            output_shape.append(1)
    return tuple(output_shape)


def mesh_pwc_durations(pwcs):
    """
    Return an array with the durations resulting of meshing the durations
    of the input PWC functions.

    Parameters
    ----------
    pwcs : list[Pwc or SparsePwc]
        The PWC functions whose durations should be meshed.

    Returns
    -------
    np.array
        The array with meshed durations.
    """

    times = np.unique(np.concatenate([np.cumsum(pwc.durations) for pwc in pwcs]))
    return np.diff(np.insert(times, 0, 0))


def validate_filter_function_input(
    control_hamiltonian, noise_operator, frequencies, sample_count, projection_operator
):
    """
    Validate the input of a filter function calculation and
    return the dimension of a Hamiltonian subject to noise.

    Parameters
    ----------
    control_hamiltonian : Pwc
        The control Hamiltonian,
    noise_operator : Pwc
        The noise operator,
    frequencies : np.ndarray
        The frequencies.
    sample_count : int, optional
        The number of samples for the Fourier transform.
    projection_operator : np.ndarray, optional
        The projection operator.

    Returns
    -------
    int
        The dimension of the system.
    """

    validate_hamiltonian(control_hamiltonian, "control_hamiltonian")
    batch_shape, value_shape = validate_batch_and_value_shapes(
        control_hamiltonian, "control_hamiltonian"
    )
    check_argument(
        len(batch_shape) == 0,
        "Batches in control Hamiltonian are not supported.",
        {"control_hamiltonian": control_hamiltonian},
    )
    batch_shape_noise, value_shape_noise = validate_batch_and_value_shapes(
        noise_operator, "noise_operator"
    )
    check_argument(
        len(batch_shape_noise) == 0,
        "Batches in noise operator are not supported.",
        {"noise_operator": noise_operator},
    )
    check_argument(
        value_shape_noise == value_shape,
        "Control Hamiltonian and noise operator must have same value shape.",
        {"control_hamiltonian": control_hamiltonian, "noise_operator": noise_operator},
    )
    dimension = value_shape[-1]

    check_argument_real_vector(frequencies, "frequencies")
    if sample_count is not None:
        check_argument_integer(sample_count, "sample_count")
        check_argument(
            sample_count > 1,
            "If provided, the number of samples must be larger than one.",
            {"sample_count": sample_count},
        )
    if projection_operator is not None:
        check_numeric_numpy_array(projection_operator, "projection_operator")
        check_argument_orthogonal_projection_operator(
            projection_operator, "projection_operator"
        )
        projection_operator_shape = validate_shape(
            projection_operator, "projection_operator"
        )
        check_argument(
            projection_operator_shape == (dimension, dimension),
            "The projection operator must have the same dimension as the control Hamiltonian.",
            {
                "projection_operator": projection_operator,
                "control_hamiltonian": control_hamiltonian,
            },
        )

    return dimension


def check_square_pwc_or_stf(pwc_or_stf, pwc_or_stf_name):
    """
    Checks that the Pwc or Stf is a square matrix, or batch of square matrices.

    Parameters
    ----------
    pwc_or_stf : Pwc or Stf
        The pwc_or_stf to be tested.
    pwc_or_stf_name : str
        The name of the `pwc_or_stf`, used for error messages.
    """
    check_argument(
        isinstance(pwc_or_stf, (Pwc, Stf)),
        f"The {pwc_or_stf_name} must be a Pwc or an Stf.",
        {pwc_or_stf_name: pwc_or_stf},
    )

    check_argument(
        len(pwc_or_stf.value_shape) == 2,
        f"The value of {pwc_or_stf_name} must be a matrix, or batch of matrices.",
        {pwc_or_stf_name: pwc_or_stf},
    )
    check_argument(
        pwc_or_stf.value_shape[0] == pwc_or_stf.value_shape[1],
        f"The value of {pwc_or_stf_name} must be a square matrix, or batch of square matrices.",
        {pwc_or_stf_name: pwc_or_stf},
    )


def check_optimization_variable_parameters(
    initial_values, count, lower_bound, upper_bound
):
    """
    Checks that the initial values for optimization variables are valid.

    Parameters
    ----------
    initial_values : list[np.ndarray] or np.ndarray
        Initial values to be checked.
    count : int
        The number of optimization variables.
    lower_bound : float
        The lower bound of initial values.
    upper_bound : float
        The upper bound of initial values.
    """
    check_argument(count > 0, "count must be positive.", {"count": count})
    check_argument(
        upper_bound > lower_bound,
        "lower_bound must be less than upper_bound.",
        {"lower_bound": lower_bound, "upper_bound": upper_bound},
    )

    def _check_value_array(value_array):
        check_argument(
            isinstance(value_array, np.ndarray) and len(value_array.shape) == 1,
            "Initial values must be a 1D array or a list of 1D arrays.",
            {"initial_values": initial_values, "upper bound": upper_bound},
        )
        check_argument(
            value_array.shape[0] == count,
            "Initial values must have the same length as count.",
            {"initial_values": initial_values, "count": count},
        )

        check_argument(
            np.all(value_array >= lower_bound),
            "Initial values must not be smaller than the lower bound.",
            {"initial_values": initial_values, "lower_bound": lower_bound},
        )
        check_argument(
            np.all(value_array <= upper_bound),
            "Initial values must not be greater than the upper bound.",
            {"initial_values": initial_values, "upper_bound": upper_bound},
        )

    if initial_values is not None:
        if isinstance(initial_values, list):
            for value in initial_values:
                _check_value_array(value)
        else:
            _check_value_array(initial_values)


def check_sample_times_with_bounds(sample_times, sample_times_name, pwc, pwc_name):
    """
    Checks that the sample_times array is valid (see check_sample_times) and
    that its values lie between 0 and sum(pwc.durations) (included).

    Parameters
    ----------
    sample_times : np.ndarray
        The array to be tested.
    sample_times_name : str
        The name of the array.
    pwc : Union[Pwc, SparsePwc]
        The PWC function whose duration bounds the values of the sample times.
    pwc_name : str
        The name of the PWC function.
    """
    check_sample_times(sample_times, sample_times_name)

    duration = np.sum(pwc.durations)
    check_argument(
        (sample_times[0] >= 0.0 or np.isclose(sample_times[0], 0.0))
        and (sample_times[-1] <= duration or np.isclose(sample_times[-1], duration)),
        f"{sample_times_name} must be between 0 and the duration of {pwc_name}.",
        {sample_times_name: sample_times, pwc_name: pwc},
        extras={"sum({duration_pwc_name}.durations)": duration},
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from dataclasses import dataclass
from enum import Enum
from typing import Any

from boulderopal._validation.basic import validate_enum


class _VerbosityEnum(Enum):
    """
    Verbosity modes for Boulder Opal cloud calculations.
    """

    VERBOSE = "VERBOSE"
    QUIET = "QUIET"


def _format_message(action_id: str, status: str) -> str:
    _message = {
        "PENDING": "is queued",
        "RECEIVED": "has been received",
        "STARTED": "has started",
        "SUCCESS": "has completed",
        "FAILURE": "has failed",
        "RETRY": "is retrying",
        "REVOKED": "has been cancelled",
    }.get(status, f"has an unknown status ({status})")

    return f'Your task (action_id="{action_id}") {_message}.'


@dataclass
class StatusMessagePrinter:
    """
    Class to manage printing messages during task execution.

    Parameters
    ----------
    last_message : str
        The last message printed.
    verbosity : str
        The level of verbosity for messages.
    """

    last_message: str = ""
    verbosity: str = "VERBOSE"

    def __call__(self, event: str, data: Any) -> None:
        """
        Prints a message if the action was updated.
        """
        if self.verbosity == "QUIET":
            return

        if event == "action.updated":
            action = data["action"]
            message = _format_message(action.action_id, action.status)
            if message != self.last_message:
                self.last_message = message
                print(message)


MESSAGE_PRINTER = StatusMessagePrinter()


def set_verbosity(verbosity: str) -> None:
    """
    Set the verbosity mode when running calculations on the cloud.

    Parameters
    ----------
    verbosity : str
        The verbosity of messages. Possible values are:
        "VERBOSE" (showing task status messages) and
        "QUIET" (not showing them).
    """
    MESSAGE_PRINTER.verbosity = validate_enum(_VerbosityEnum, verbosity)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


from __future__ import annotations

from typing import Optional

from boulderopal._core import run_gradient_free_optimization_workflow
from boulderopal._validation.basic import (
    ScalarDType,
    nullable,
    validate_enum,
)
from boulderopal._validation.graph import (
    check_cost_node,
    check_cost_node_for_optimization_graph,
    check_initial_value_for_optimization_node,
    check_optimization_node_in_graph,
    validate_output_node_names,
)
from boulderopal.graph._graph import Graph
from boulderopal.optimization._optimization import HistoryScope


def run_gradient_free_optimization(
    graph: Graph,
    cost_node_name: str,
    output_node_names: str | list[str],
    iteration_count: int = 100,
    target_cost: Optional[float] = None,
    optimization_count: int = 4,
    cost_history_scope: HistoryScope = HistoryScope.NONE,
    seed: Optional[int] = None,
) -> dict:
    r"""
    Perform model-based optimization without using gradient values.

    Use this function to determine a choice of variables that minimizes the
    value of a scalar real-valued cost function of those variables. You
    express that cost function as a graph describing how the input variables
    are transformed into the cost value.

    This function provides an alternative to the gradient based optimizer, and
    is useful when the gradient is either very costly to compute or inaccessible
    (for example if the graph includes a node that does not allow gradients).

    Parameters
    ----------
    graph : Graph
        The graph describing the cost :math:`C(\mathbf v)` and outputs :math:`\{F_j(\mathbf v)\}`
        as functions of the optimizable input variables :math:`\mathbf v`.
        The graph must contain nodes with names :math:`s` (giving the cost function)
        and :math:`\{s_j\}` (giving the output functions).
    cost_node_name : str
        The name :math:`s` of the real-valued scalar graph node that defines the cost
        function :math:`C(\mathbf v)` to be minimized.
    output_node_names : str or list[str]
        The names :math:`\{s_j\}` of the graph nodes that define the output functions
        :math:`\{F_j(\mathbf v)\}`. The function evaluates these using the optimized
        variables and returns them in the output.
    iteration_count : int, optional
        The number of iterations the optimizer performs until it halts.
        Defaults to 100.
    target_cost : float or None, optional
        A target value of the cost that you can set as an early stop condition for the optimizer.
        If the cost becomes equal or smaller than this value, the optimization halts.
        Defaults to None, meaning that the optimizer runs until it converges.
    optimization_count : int, optional
        The number :math:`N` of independent randomly seeded optimizations to perform.
        The function returns the results from the best optimization (the one with
        the lowest cost). Defaults to 4. Depending on the landscape of the optimization problem,
        a larger value will help in finding lower costs, at the expense of prolonging computation
        time.
    cost_history_scope : HistoryScope, optional
        Configuration for the scope of the returned cost history data.
        Use this to select if you want the history data to be returned.
        Defaults to no cost history data returned.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.
        Note that if your graph contains nodes generating random samples,
        you need to also set seeds for those nodes to ensure a reproducible result.

    Returns
    -------
    dict
        A dictionary containing the optimization result, with the following keys:

        ``cost``
            The minimum cost function value :math:`C(\mathbf v_\mathrm{optimized})`
            achieved across all optimizations.
        ``output``
            The dictionary giving the value of each requested output node, evaluated
            at the optimized variables, namely
            :math:`\{s_j: F_j(\mathbf v_\mathrm{optimized})\}`.
            The keys of the dictionary are the names :math:`\{s_j\}` of the output nodes.
        ``cost_history``
            The evolution of the cost function across all optimizations and iterations.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    boulderopal.closed_loop.optimize :
        Run a closed-loop optimization to find a minimum of the given cost function.
    boulderopal.closed_loop.step : Perform a single step in a closed-loop optimization.
    boulderopal.execute_graph : Evaluate generic functions.
    boulderopal.run_optimization :
        Perform gradient-based deterministic optimization of generic real-valued functions.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.

    Notes
    -----
    Given a cost function :math:`C(\mathbf v)` of variables :math:`\mathbf v`, this function
    computes an estimate :math:`\mathbf v_\mathrm{optimized}` of
    :math:`\mathrm{argmin}_{\mathbf v} C(\mathbf v)`, namely the choice of variables
    :math:`\mathbf v` that minimizes :math:`C(\mathbf v)`. The function then calculates the values
    of arbitrary output functions :math:`\{F_j(\mathbf v_\mathrm{optimized})\}` with that choice
    of variables.

    This function represents the cost and output functions as nodes of a graph.
    This graph defines the input optimization variables :math:`\mathbf v`, and how these variables
    are transformed into the corresponding cost and output quantities. You build the graph from
    primitive nodes defined in the graph of the Boulder Opal package.
    Each such node, which can be identified by a name, represents a function of the previous nodes
    in the graph (and thus, transitively, a function of the input variables).
    You can use any named scalar real-valued node :math:`s` as the cost function, and any named
    nodes :math:`\{s_j\}` as outputs.

    After you provide a cost function :math:`C(\mathbf v)` (via a graph), this function runs an
    optimization process :math:`N` times, each with random initial variables, to identify :math:`N`
    local minima of the cost function, and then takes the variables corresponding to the best
    such minimum as :math:`\mathbf v_\mathrm{optimized}`.

    The optimizer will carry out the gradient-free optimization for a number of iterations
    set by `iteration_count`. If a `target_cost` is passed and the cost becomes less
    than or equal to this value then the optimizer will terminate early.

    The gradient-free optimizer used here is the covariance matrix adaptation evolution strategy
    (CMA-ES) based optimizer. For more detail on CMA-ES see
    `CMA-ES <https://en.wikipedia.org/wiki/CMA-ES>`_ on Wikipedia.

    Examples
    --------
    Perform a simple optimization with variables that are initialized to given values.

    >>> graph = bo.Graph()
    >>> variables = graph.optimization_variable(
    ...     2, -1, 1, initial_values=np.array([0.6, 0.]), name="variables"
    ... )
    >>> x, y = variables[0], variables[1]
    >>> cost = (x - 0.5) ** 2 + (y - 0.1) ** 2
    >>> cost.name = "cost"
    >>> result = bo.run_gradient_free_optimization(
    ...     graph=graph, cost_node_name="cost", output_node_names="variables", seed=0
    ... )
    >>> result["cost"], result["output"]
        (5.166836148557636e-22, {'variables': {'value': array([0.5, 0.1])}})

    See also the `How to optimize controls using gradient-free optimization user guide
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-
    optimize-controls-using-gradient-free-optimization>`_.
    """
    check_optimization_node_in_graph(graph)
    check_initial_value_for_optimization_node(graph)
    check_cost_node(cost_node_name, graph)

    output_node_names = validate_output_node_names(output_node_names, graph)
    check_cost_node_for_optimization_graph(
        graph, cost_node_name, output_node_names, check_gradient_nodes=False
    )

    result = run_gradient_free_optimization_workflow(
        graph=graph,
        cost_node_name=cost_node_name,
        output_node_names=output_node_names,
        iteration_count=nullable(
            ScalarDType.INT, iteration_count, "iteration_count", min_=0
        ),
        target_cost=nullable(ScalarDType.REAL, target_cost, "target_cost"),
        optimization_count=ScalarDType.INT(
            optimization_count, "optimization_count", min_=0
        ),
        cost_history_scope=validate_enum(HistoryScope, cost_history_scope),
        seed=nullable(ScalarDType.INT, seed, "seed", min_=0, min_inclusive=True),
    )
    return result



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from ._optimization import (
    Results,
    optimize,
    step,
)
from ._optimizers import (
    Bounds,
    BoundType,
    ClosedLoopOptimizer,
    Cmaes,
    GaussianProcess,
    NeuralNetwork,
    SimulatedAnnealing,
)

__all__ = [
    "Bounds",
    "BoundType",
    "ClosedLoopOptimizer",
    "Cmaes",
    "GaussianProcess",
    "NeuralNetwork",
    "Results",
    "SimulatedAnnealing",
    "optimize",
    "step",
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.composite.namespace import composite_node
from boulderopal._nodes.documentation import Category
from boulderopal._nodes.node_data import Pwc


@composite_node([Category.OPTIMIZATION_VARIABLES])
def real_optimizable_pwc_signal(
    graph: "Graph",
    segment_count: int,
    duration: float,
    maximum: float,
    minimum: float = 0.0,
    initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
    name: Optional[str] = None,
) -> Pwc:
    """
    Create a real optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of piecewise-constant segments in the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the signal values.
    minimum : float, optional
        The lower bound for the signal values. Defaults to 0.
    initial_values : np.ndarray or list[np.ndarray] or None, optional
        Initial values for the signal. You can either provide a single array,
        or a list of them. If a list of arrays is used, they must have the same length.
        Defaults to None, meaning the optimizer initializes the variables with random values.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`Graph.complex_optimizable_pwc_signal` :
        Create a complex optimizable `Pwc` signal.
    :func:`Graph.optimization_variable` :
        Create optimization variables, which can be bounded, semi-bounded, or unbounded.
    :func:`Graph.pwc_signal` : Create a piecewise-constant signal.
    """

    if initial_values is not None:
        check_argument(
            np.all(np.isreal(initial_values)),
            "Initial signal values must be real.",
            {"initial_values": initial_values},
        )

    values = graph.optimization_variable(
        count=segment_count,
        lower_bound=minimum,
        upper_bound=maximum,
        initial_values=initial_values,
    )
    return graph.pwc_signal(values=values, duration=duration, name=name)


@composite_node([Category.OPTIMIZATION_VARIABLES])
def complex_optimizable_pwc_signal(
    graph: "Graph",
    segment_count: int,
    duration: float,
    maximum: float,
    initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
    name: Optional[str] = None,
) -> Pwc:
    """
    Create a complex optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of segments of the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the modulus of the signal values.
    initial_values : np.ndarray or list[np.ndarray] or None, optional
        Initial values for the signal. You can either provide a single array,
        or a list of them. If a list of arrays is used, they must have the same length.
        Defaults to None, meaning the optimizer initializes the variables with random values.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`Graph.real_optimizable_pwc_signal` :
        Create a real optimizable `Pwc` signal.
    :func:`Graph.complex_pwc_signal` :
        Create a complex piecewise-constant signal from moduli and phases.
    :func:`Graph.optimization_variable` :
        Create optimization variables, which can be bounded, semi-bounded, or unbounded.
    :func:`Graph.pwc_signal` : Create a piecewise-constant signal.

    Notes
    -----
    Note that this function sets limits to the modulus of the signal.

    If you want to set (different) limits to the real and imaginary parts instead,
    consider using `graph.real_optimizable_signal` to create signals for the
    real and imaginary parts, which you can pass to `graph.complex_value`.
    """

    initial_moduli: Optional[list[np.ndarray] | np.ndarray] = None
    initial_phases: Optional[list[np.ndarray] | np.ndarray] = None

    if initial_values is not None:
        if isinstance(initial_values, list):
            initial_moduli = list(np.absolute(initial_values))
            initial_phases = list(np.angle(initial_values))
        else:
            initial_moduli = np.absolute(initial_values)
            initial_phases = np.angle(initial_values)

    moduli = graph.optimization_variable(
        count=segment_count,
        lower_bound=0.0,
        upper_bound=maximum,
        initial_values=initial_moduli,
    )
    phases = graph.optimization_variable(
        count=segment_count,
        lower_bound=-np.pi,
        upper_bound=np.pi,
        is_lower_unbounded=True,
        is_upper_unbounded=True,
        initial_values=initial_phases,
    )
    return graph.complex_pwc_signal(
        moduli=moduli, phases=phases, duration=duration, name=name
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
STF signal library nodes.
"""

from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.composite.signal_utils import validate_optimizable_parameter
from boulderopal._nodes.node_data import (
    Stf,
    Tensor,
)


class StfSignals:
    """
    Class for the STF Signal nodes.
    """

    def __init__(self, graph):
        self._graph = graph

    def sech_pulse_stf(
        self,
        amplitude: float | complex | Tensor,
        width: float | Tensor,
        center_time: float | Tensor,
    ) -> Stf:
        # pylint: disable=line-too-long
        r"""
        Create an `Stf` representing a hyperbolic secant pulse.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor
            The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
            It must either be a scalar or contain a single element.
        center_time : float or Tensor
            The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
            It must either be a scalar or contain a single element.

        Returns
        -------
        Stf
            The sampleable hyperbolic secant pulse.

        See Also
        --------
        :func:`Graph.signals.gaussian_pulse_stf <signals.gaussian_pulse_stf>`
            Create an `Stf` representing a Gaussian pulse.
        :func:`boulderopal.signals.sech_pulse`
            Create a `Signal` object representing a hyperbolic secant pulse.
        :func:`Graph.signals.sech_pulse_pwc <signals.sech_pulse_pwc>`
            Corresponding operation with `Pwc` output.

        Notes
        -----
        The hyperbolic secant pulse is defined as

            .. math:: \mathop{\mathrm{Sech}}(t)
                = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} .

        The full width at half maximum of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

        Examples
        --------
        Define a sampleable sech pulse.

        >>> sech = graph.signals.sech_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5
        ... )
        >>> sech
        <Stf: operation_name="truediv", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=sech, sample_times=np.linspace(0, 1, 5), name="sech_samples")
        <Tensor: name="sech_samples", operation_name="sample_stf", shape=(5,)>
        >>> graph.discretize_stf(stf=sech, duration=1.2, segment_count=100, name="discretized_sech")
        <Pwc: name="discretized_sech", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["sech_samples", "discretized_sech"]
        ... )
        >>> result["output"]["sech_samples"]["value"]
        array([0.013, 0.163, 1.000, 0.163, 0.013])

        Define a sampleable sech pulse with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> width = graph.optimization_variable(
        ...     count=1, lower_bound=0.1, upper_bound=0.5, name="width"
        ... )
        >>> center_time = graph.optimization_variable(
        ...     count=1, lower_bound=0.2, upper_bound=0.8, name="center_time"
        ... )
        >>> graph.signals.sech_pulse_stf(
        ...     amplitude=amplitude, width=width, center_time=center_time
        ... )
        <Stf: operation_name="truediv", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        width = validate_optimizable_parameter(graph, width, "width")
        center_time = validate_optimizable_parameter(graph, center_time, "center time")

        return amplitude / graph.cosh((graph.identity_stf() - center_time) / width)

    def gaussian_pulse_stf(
        self,
        amplitude: float | complex | Tensor,
        width: float | Tensor,
        center_time: float | Tensor,
        drag: Optional[float | Tensor] = None,
    ) -> Stf:
        # pylint: disable=line-too-long
        r"""
        Create an `Stf` representing a Gaussian pulse.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the Gaussian pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor
            The standard deviation of the Gaussian pulse, :math:`\sigma`.
            It must either be a scalar or contain a single element.
        center_time : float or Tensor
            The center of the Gaussian pulse, :math:`t_0`.
            It must either be a scalar or contain a single element.
        drag : float or Tensor or None, optional
            The DRAG parameter, :math:`\beta`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to no DRAG correction.

        Returns
        -------
        Stf
            The sampleable Gaussian pulse.

        See Also
        --------
        :func:`Graph.signals.gaussian_pulse_pwc <signals.gaussian_pulse_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.sech_pulse_stf <signals.sech_pulse_stf>`
            Create an `Stf` representing a hyperbolic secant pulse.

        Notes
        -----
        The Gaussian pulse is defined as

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
            A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) .

        Examples
        --------
        Define a sampleable Gaussian pulse.

        >>> gaussian = graph.signals.gaussian_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5
        ... )
        >>> gaussian
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=gaussian, sample_times=np.linspace(0, 1, 5), name="gaussian_samples")
        <Tensor: name="gaussian_samples", operation_name="sample_stf", shape=(5,)>
        >>> graph.discretize_stf(
        ... gaussian, duration=1, segment_count=100, name="discretized_gaussian"
        ... )
        <Pwc: name="discretized_gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["gaussian_samples", "discretized_gaussian"]
        ... )
        >>> result["output"]["gaussian_samples"]["value"]
        array([3.727e-06, 4.394e-02, 1.000e+00, 4.394e-02, 3.727e-06])

        Define a sampleable Gaussian with a DRAG correction.

        >>> gaussian = graph.signals.gaussian_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5, drag=0.2
        ... )
        >>> gaussian
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(
        ...     stf=gaussian, sample_times=np.linspace(0, 1, 5), name="drag_gaussian_samples"
        ... )
        <Tensor: name="drag_gaussian_samples", operation_name="sample_stf", shape=(5,)>
        >>> graph.discretize_stf(
        ...     gaussian, duration=1, segment_count=100, name="discretized_drag_gaussian"
        ... )
        <Pwc: name="discretized_drag_gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["drag_gaussian_samples", "discretized_drag_gaussian"]
        ... )
        >>> result["output"]["drag_gaussian_samples"]["value"]
        array([
            3.727e-06 + 9.317e-06j,
            4.394e-02 + 5.492e-02j,
            1.000e+00 + 0.000e+00j,
            4.394e-02 - 5.492e-02j,
            3.727e-06 - 9.317e-06j,
        ])

        Define a sampleable Gaussian pulse with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> width = graph.optimization_variable(
        ...     count=1, lower_bound=0.1, upper_bound=0.5, name="width"
        ... )
        >>> center_time = graph.optimization_variable(
        ...     count=1, lower_bound=0.2, upper_bound=0.8, name="center_time"
        ... )
        >>> drag = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=0.5, name="drag"
        ... )
        >>> graph.signals.gaussian_pulse_stf(
        ...     amplitude=amplitude, width=width, center_time=center_time, drag=drag
        ... )
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long
        graph = self._graph

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        width = validate_optimizable_parameter(graph, width, "width")
        center_time = validate_optimizable_parameter(graph, center_time, "center time")

        if drag is not None:
            drag = validate_optimizable_parameter(graph, drag, "drag")
            correction = -(1j * drag / (width**2)) * (
                graph.identity_stf() - center_time
            )
            amplitude *= 1.0 + correction

        return amplitude * graph.exp(
            -((graph.identity_stf() - center_time) ** 2) / (2 * width**2)
        )

    def sinusoid_stf(
        self,
        amplitude: float | complex | Tensor,
        angular_frequency: float | Tensor,
        phase: float | Tensor = 0.0,
    ) -> Stf:
        # pylint: disable=line-too-long
        r"""
        Create an `Stf` representing a sinusoidal oscillation.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the oscillation, :math:`A`.
            It must either be a scalar or contain a single element.
        angular_frequency : float or Tensor
            The angular frequency of the oscillation, :math:`\omega`.
            It must either be a scalar or contain a single element.
        phase : float or Tensor, optional
            The phase of the oscillation, :math:`\phi`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable sinusoid.

        See Also
        --------
        :func:`Graph.signals.hann_series_stf <signals.hann_series_stf>`
            Create an `Stf` representing a sum of Hann window functions.
        :func:`boulderopal.signals.sinusoid`
            Create a `Signal` object representing a sinusoidal oscillation.
        :func:`Graph.signals.sinusoid_pwc <signals.sinusoid_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.sin`
            Calculate the element-wise sine of an object.

        Notes
        -----
        The sinusoid is defined as

        .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) .

        Examples
        --------
        Define an STF oscillation.

        >>> oscillation = graph.signals.sinusoid_stf(
        ...     amplitude=1.0, angular_frequency=np.pi, phase=np.pi/2.0
        ... )
        >>> oscillation
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=oscillation, sample_times=np.linspace(0, 1, 5), name="oscillation")
        <Tensor: name="oscillation", operation_name="sample_stf", shape=(5,)>
        >>> graph.discretize_stf(
        ...     oscillation, duration=10, segment_count=100, name="discretized_oscillation"
        ... )
        <Pwc: name="discretized_oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["oscillation", "discretized_oscillation"]
        ... )
        >>> result["output"]["oscillation"]["value"]
        array([ 1.000e+00,  7.071e-01,  1.225e-16, -7.071e-01, -1.000e+00])

        Define a sinusoid with optimizable parameters.

        >>> amplitude = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=4e3, name="amplitude"
        ... )
        >>> angular_frequency = graph.optimization_variable(
        ...     count=1, lower_bound=5e6, upper_bound=20e6, name="angular_frequency"
        ... )
        >>> phase = graph.optimization_variable(
        ...     count=1,
        ...     lower_bound=0,
        ...     upper_bound=2*np.pi,
        ...     is_lower_unbounded=True,
        ...     is_upper_unbounded=True,
        ...     name="phase",
        ... )
        >>> graph.signals.sinusoid_stf(
        ...     amplitude=amplitude, angular_frequency=angular_frequency, phase=phase
        ... )
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long
        graph = self._graph

        amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
        angular_frequency = validate_optimizable_parameter(
            graph, angular_frequency, "angular frequency"
        )
        phase = validate_optimizable_parameter(graph, phase, "phase")

        return amplitude * graph.sin(angular_frequency * graph.identity_stf() + phase)

    def hann_series_stf(
        self,
        coefficients: np.ndarray | Tensor,
        end_time: float,
        start_time: float = 0.0,
    ) -> Stf:
        # pylint: disable=line-too-long
        r"""
        Create an `Stf` representing a sum of Hann window functions.

        Parameters
        ----------
        coefficients : np.ndarray or Tensor
            The coefficients for the different Hann window functions, :math:`c_n`.
            It must be a 1D array or Tensor.
        end_time : float
            The time at which the Hann series ends, :math:`t_\mathrm{end}`.
        start_time : float, optional
            The time at which the Hann series starts, :math:`t_\mathrm{start}`.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable Hann window functions series.

        See Also
        --------
        :func:`boulderopal.signals.hann_series`
            Create a `Signal` object representing a sum of Hann window functions.
        :func:`Graph.signals.hann_series_pwc <signals.hann_series_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.sinusoid_stf <signals.sinusoid_stf>`
            Create an `Stf` representing a sinusoidal oscillation.

        Notes
        -----
        The series is defined as

        .. math:: \mathop{\mathrm{Hann}}(t)
            = \sum_{n=1}^N c_n \sin^2 \left(
                \frac{\pi n (t - t_\mathrm{start})}{t_\mathrm{end} - t_\mathrm{start}}
            \right) ,

        where :math:`N` is the number of coefficients.

        Note that the function values outside the :math:`(t_\mathrm{start}, t_\mathrm{end})` range
        will not be zero.

        Examples
        --------
        Define a simple sampleable Hann series.

        >>> hann = graph.signals.hann_series_stf(coefficients=np.array([0.5, 1, 0.25]), end_time=1.0)
        >>> hann
        <Stf: operation_name="stf_sum", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=hann, sample_times=np.linspace(0, 1, 7), name="hann_samples")
        <Tensor: name="hann_samples", operation_name="hann_samples", shape=(7,)>
        >>> graph.discretize_stf(hann, duration=1, segment_count=100, name="discretized_hann")
        <Pwc: name="discretized_hann", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["hann_samples", "discretized_hann"]
        ... )
        >>> result["output"]["hann_samples"]["value"]
        array([0.000e+00, 1.125e+00, 1.125e+00, 7.500e-01, 1.125e+00, 1.125e+00, 5.159e-14])

        Define a sampleable Hann series with optimizable coefficients.

        >>> coefficients = graph.optimization_variable(
        ...     count=8, lower_bound=-3.5e6, upper_bound=3.5e6, name="coefficients"
        ... )
        >>> graph.signals.hann_series_stf(coefficients=coefficients, end_time=2.0e-6)
        <Stf: operation_name="stf_sum", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        check_argument(
            len(coefficients.shape) == 1,
            "The coefficients must be in a 1D array or Tensor.",
            {"coefficients": coefficients},
            extras={"coefficients.shape": coefficients.shape},
        )

        check_argument(
            end_time > start_time,
            "The end time must be greater than the start time.",
            {"start_time": start_time, "end_time": end_time},
        )

        # Define scaled times π (t - t_start) / (t_end - t_start).
        scaled_time = (graph.identity_stf() - start_time) * (
            np.pi / (end_time - start_time)
        )

        # Calculate function values.
        stfs = [
            coefficients[idx] * graph.sin((idx + 1) * scaled_time) ** 2
            for idx in range(coefficients.shape[0])
        ]
        return graph.stf_sum(stfs)

    def linear_ramp_stf(
        self, slope: float | complex | Tensor, shift: float | complex | Tensor = 0.0
    ) -> Stf:
        # pylint:disable=line-too-long
        r"""
        Create an `Stf` representing a linear ramp.

        Parameters
        ----------
        slope : float or complex or Tensor
            The slope of the ramp, :math:`a`.
            It must either be a scalar or contain a single element.
        shift : float or complex or Tensor, optional
            The value of the ramp at :math:`t = 0`, :math:`b`.
            It must either be a scalar or contain a single element.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable linear ramp.

        See Also
        --------
        :func:`boulderopal.signals.linear_ramp`
            Create a `Signal` object representing a linear ramp.
        :func:`Graph.signals.linear_ramp_pwc <signals.linear_ramp_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.tanh_ramp_stf <signals.tanh_ramp_stf>`
            Create an `Stf` representing a hyperbolic tangent ramp.

        Notes
        -----
        The linear ramp is defined as

        .. math:: \mathop{\mathrm{Linear}}(t) = a t + b .

        Examples
        --------
        Define a linear STF ramp.

        >>> linear = graph.signals.linear_ramp_stf(slope=4.0, shift=-2.0)
        >>> linear
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=linear, sample_times=np.linspace(0, 1, 5), name="linear_ramp")
        <Tensor: name="linear_ramp", operation_name="sample_stf", shape=(5,)>
        >>> graph.discretize_stf(linear, duration=1, segment_count=100, name="discretized_linear")
        <Pwc: name="discretized_linear", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["linear_ramp", "discretized_linear"]
        ... )
        >>> result["output"]["linear_ramp"]["value"]
        array([-2., -1., 0., 1., 2.]

        Define a linear STF ramp with an optimizable slope and root.

        >>> slope = graph.optimization_variable(
        ...     count=1, lower_bound=-4, upper_bound=4, name="slope"
        ... )
        >>> root = graph.optimization_variable(
        ...     count=1, lower_bound=-4, upper_bound=4, name="slope"
        ... )
        >>> shift = - slope * root
        >>> graph.signals.linear_ramp_stf(slope=slope, shift=shift)
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        """
        # pylint:enable=line-too-long
        graph = self._graph

        slope = validate_optimizable_parameter(graph, slope, "slope")
        shift = validate_optimizable_parameter(graph, shift, "shift")

        return slope * graph.identity_stf() + shift

    def tanh_ramp_stf(
        self,
        center_time: float | Tensor,
        ramp_duration: float | Tensor,
        end_value: float | complex | Tensor,
        start_value: Optional[float | complex | Tensor] = None,
    ) -> Stf:
        # pylint: disable=line-too-long
        r"""
        Create an `Stf` representing a hyperbolic tangent ramp.

        Parameters
        ----------
        center_time : float or Tensor
            The time at which the ramp has its greatest slope, :math:`t_0`.
            It must either be a scalar or contain a single element.
        ramp_duration : float or Tensor
            The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
            It must either be a scalar or contain a single element.
        end_value : float or complex or Tensor
            The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
            It must either be a scalar or contain a single element.
        start_value : float or complex or Tensor, optional
            The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to minus `end_value`.

        Returns
        -------
        Stf
            The sampleable hyperbolic tangent ramp.

        See Also
        --------
        :func:`Graph.signals.linear_ramp_stf <signals.linear_ramp_stf>`
            Create an `Stf` representing a linear ramp.
        :func:`boulderopal.signals.tanh_ramp`
            Create a `Signal` object representing a hyperbolic tangent ramp.
        :func:`Graph.signals.tanh_ramp_pwc <signals.tanh_ramp_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.tanh`
            Calculate the element-wise hyperbolic tangent of an object.

        Notes
        -----
        The hyperbolic tangent ramp is defined as

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = \frac{a_+ + a_-}{2}
                + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

        where the function's asymptotic values :math:`a_\pm` are defined by:

        .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) ,

        and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

        .. math::
            \left.\frac{{\rm d}\mathop{\mathrm{Tanh}}(t)}{{\rm d}t}\right|_{t=t_0}
                = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} .

        With the default value of `start_value` (:math:`a_-`),
        the ramp expression simplifies to

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = A \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

        where :math:`A = a_+` is the end value (the start value is then :math:`-A`).

        Examples
        --------
        Define a simple sampleable hyperbolic tangent ramp.

        >>> tanh = graph.signals.tanh_ramp_stf(
        ...     center_time=0.4, ramp_duration=0.2, end_value=2, start_value=-1
        ... )
        >>> tanh
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        >>> graph.sample_stf(stf=tanh, sample_times=np.linspace(0, 1, 7), name="tanh_samples")
        <Tensor: name="tanh_samples", operation_name="sample_stf", shape=(7,)>
        >>> graph.discretize_stf(tanh, duration=1, segment_count=100, name="discretized_tanh")
        <Pwc: name="discretized_tanh", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(
        ...     graph=graph, output_node_names=["tanh_samples", "discretized_tanh"]
        ... )
        >>> result["output"]["tanh_samples"]["value"]
        array([-0.946, -0.735,  0.018,  1.193,  1.805,  1.961,  1.993])

        Define a hyperbolic tangent ramp with optimizable parameters.

        >>> center_time = graph.optimization_variable(
        ...     count=1, lower_bound=0.25e-6, upper_bound=0.75e-6, name="center_time"
        ... )
        >>> ramp_duration = graph.optimization_variable(
        ...     count=1, lower_bound=0.1e-6, upper_bound=0.3e-6, name="ramp_duration"
        ... )
        >>> end_value = graph.optimization_variable(
        ...     count=1, lower_bound=0, upper_bound=3e6, name="end_value"
        ... )
        >>> graph.signals.tanh_ramp_stf(
        ...     center_time=center_time, ramp_duration=ramp_duration, end_value=end_value
        ... )
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        """
        # pylint: enable=line-too-long

        graph = self._graph

        if start_value is None:
            start_value = -end_value

        center_time = validate_optimizable_parameter(graph, center_time, "center time")
        ramp_duration = validate_optimizable_parameter(
            graph, ramp_duration, "ramp duration"
        )
        end_value = validate_optimizable_parameter(graph, end_value, "end value")
        start_value = validate_optimizable_parameter(graph, start_value, "start value")

        return start_value + 0.5 * (end_value - start_value) * (
            1 + graph.tanh((graph.identity_stf() - center_time) / ramp_duration)
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from boulderopal.signals._functions import (
    Signal,
    cosine_pulse,
    gaussian_pulse,
    hann_series,
    linear_ramp,
    sech_pulse,
    sinusoid,
    square_pulse,
    tanh_ramp,
)

__all__ = [
    "Signal",
    "cosine_pulse",
    "gaussian_pulse",
    "hann_series",
    "linear_ramp",
    "sech_pulse",
    "sinusoid",
    "square_pulse",
    "tanh_ramp",
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
PWC and STF signal library nodes.
"""
from boulderopal._nodes.composite.namespace import composite_node
from boulderopal._nodes.composite.signals_pwc import (
    PwcSignals,
    SegmentationType,
)
from boulderopal._nodes.composite.signals_stf import StfSignals
from boulderopal._nodes.documentation import Category


@composite_node([Category.SIGNALS])
class SignalsNamespace(PwcSignals, StfSignals):
    """
    Class for the PWC and STF Signal nodes.
    """

    namespace_name = "signals"
    SegmentationType = SegmentationType



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_argument_iterable,
    check_argument_real_vector,
    check_duration,
    check_numeric_numpy_array,
    check_operator,
)

from .documentation import Category
from .node_data import (
    Pwc,
    Tensor,
)
from .types import TensorLike
from .utils import (
    check_sample_times_with_bounds,
    get_broadcasted_shape,
    mesh_pwc_durations,
    validate_batch_and_value_shapes,
    validate_hamiltonian,
    validate_shape,
)


class PwcOperation(Node):
    r"""
    Create a piecewise-constant function of time.

    Parameters
    ----------
    durations : np.ndarray (1D, real)
        The durations :math:`\{\delta t_n\}` of the :math:`N` constant
        segments.
    values : np.ndarray or Tensor
        The values :math:`\{v_n\}` of the function on the constant segments.
        The dimension corresponding to `time_dimension` must be the same
        length as `durations`. To create a batch of
        :math:`B_1 \times \ldots \times B_n` piecewise-constant tensors of
        shape :math:`D_1 \times \ldots \times D_m`, provide this `values`
        parameter as an object of shape
        :math:`B_1\times\ldots\times B_n\times N\times D_1\times\ldots\times D_m`.
    time_dimension : int, optional
        The axis along `values` corresponding to time. All dimensions that
        come before the `time_dimension` are batch dimensions: if there are
        :math:`n` batch dimensions, then `time_dimension` is also :math:`n`.
        Defaults to 0, which corresponds to no batch. Note that you can
        pass a negative value to refer to the time dimension.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function of time :math:`v(t)`, satisfying
        :math:`v(t)=v_n` for :math:`t_{n-1}\leq t\leq t_n`, where
        :math:`t_0=0` and :math:`t_n=t_{n-1}+\delta t_n`. If you provide a
        batch of values, the returned `Pwc` represents a
        corresponding batch of :math:`B_1 \times \ldots \times B_n`
        functions :math:`v(t)`, each of shape
        :math:`D_1 \times \ldots \times D_m`.

    See Also
    --------
    Graph.pwc_operator : Create `Pwc` operators.
    Graph.pwc_signal : Create `Pwc` signals from (possibly complex) values.
    Graph.pwc_sum : Sum multiple `Pwc`\s.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a Hamiltonian from a piecewise-constant signal with non-uniform segment durations.

    >>> omega = graph.pwc(
    ...     values=np.array([1, 2, 3]), durations=np.array([0.1, 0.2, 0.3]), name="omega"
    ... )
    >>> omega
    <Pwc: name="omega", operation_name="pwc", value_shape=(), batch_shape=()>
    >>> sigma_z = np.array([[1, 0], [0, -1]])
    >>> hamiltonian = omega * sigma_z
    >>> hamiltonian.name = "hamiltonian"
    >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
    >>> result["output"]["hamiltonian"]
    {
        'durations': array([0.1, 0.2, 0.3]),
        'values': array([
            [[ 1.,  0.], [ 0., -1.]],
            [[ 2.,  0.], [ 0., -2.]],
            [[ 3.,  0.], [ 0., -3.]]
        ]),
        'time_dimension': 0
    }

    See more examples in the `How to simulate quantum dynamics subject to noise with graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-
    quantum-dynamics-subject-to-noise-with-graphs>`_ user guide.
    """

    name = "pwc"
    args = [
        forge.arg("durations", type=np.ndarray),
        forge.arg("values", type=TensorLike),
        forge.arg("time_dimension", type=int, default=0),
    ]

    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        durations = kwargs.get("durations")
        values = kwargs.get("values")
        time_dimension = kwargs.get("time_dimension")
        check_argument_real_vector(durations, "durations")
        check_numeric_numpy_array(values, "values")
        check_argument_integer(time_dimension, "time_dimension")
        value_shape = validate_shape(values, "values")
        check_argument(
            len(value_shape) > 0 and value_shape != (0,),
            "The `values` must be a non-empty array or Tensor.",
            {"values": values},
        )
        durations_shape = validate_shape(durations, "durations")
        check_argument(
            -len(value_shape) <= time_dimension < len(value_shape),
            "The `time_dimension` must be a valid `values` dimension.",
            {"time_dimension": time_dimension, "values": values},
        )
        check_argument(
            value_shape[time_dimension] == durations_shape[0],
            "The size of the `time_dimension` of `values` must be equal to the length"
            " of `durations`.",
            {
                "time_dimension": time_dimension,
                "values": values,
                "durations": durations,
            },
        )
        for index, duration in enumerate(durations):
            check_duration(duration, f"durations[{index}]")
        if time_dimension < 0:
            time_dimension = len(value_shape) + time_dimension
        return Pwc(
            _operation,
            value_shape=value_shape[time_dimension + 1 :],
            durations=durations,
            batch_shape=value_shape[:time_dimension],
        )


class PwcSignal(Node):
    r"""
    Create a piecewise-constant signal (scalar-valued function of time).

    Use this function to create a piecewise-constant signal in which the
    constant segments all have the same duration.

    Parameters
    ----------
    values : np.ndarray or Tensor
        The values :math:`\{\alpha_n\}` of the :math:`N` constant segments.
        These can represent either a single sequence of segment values or a
        batch of them. To create a batch of
        :math:`B_1 \times \ldots \times B_n` signals, represent these
        `values` as a tensor of shape
        :math:`B_1 \times \ldots \times B_n \times N`.
    duration : float
        The total duration :math:`\tau` of the signal.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function of time :math:`\alpha(t)`, satisfying
        :math:`\alpha(t)=\alpha_n` for :math:`t_{n-1}\leq t\leq t_n`, where
        :math:`t_n=n\tau/N` (where :math:`N` is the number of values
        in :math:`\{\alpha_n\}`). If you provide a batch of values, the
        returned `Pwc` represents a corresponding batch of
        :math:`B_1 \times \ldots \times B_n` functions :math:`\alpha(t)`.

    See Also
    --------
    Graph.complex_pwc_signal : Create complex `Pwc` signals from their moduli and phases.
    Graph.pwc : Corresponding operation with support for segments of different durations.
    Graph.pwc_operator : Create `Pwc` operators.
    Graph.pwc_sum : Sum multiple `Pwc`\s.
    Graph.symmetrize_pwc : Symmetrize `Pwc`\s.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a piecewise-constant signal with uniform segment duration.

    >>> graph.pwc_signal(duration=0.1, values=np.array([2, 3]), name="signal")
    <Pwc: name="signal", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="signal")
    >>> result["output"]["signal"]
    {
        'durations': array([0.05, 0.05]),
        'values': array([2., 3.]),
        'time_dimension': 0
    }

    See more examples in the `Get familiar with graphs <https://docs.q-ctrl.com/
    boulder-opal/tutorials/get-familiar-with-graphs>`_ tutorial.
    """

    name = "pwc_signal"
    args = [forge.arg("values", type=TensorLike), forge.arg("duration", type=float)]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        values = kwargs.get("values")
        duration = kwargs.get("duration")
        check_numeric_numpy_array(values, "values")
        check_duration(duration, "duration")
        shape = validate_shape(values, "values")
        check_argument(
            len(shape) > 0,
            "The shape of values must have at least one dimension.",
            {"values": values},
        )
        durations = duration / shape[-1] * np.ones(shape[-1])
        return Pwc(
            _operation, value_shape=(), durations=durations, batch_shape=shape[:-1]
        )


class ComplexPwcSignal(Node):
    r"""
    Create a complex piecewise-constant signal from moduli and phases.

    Use this function to create a complex piecewise-constant signal from
    moduli and phases defined for each segment, in which the constant segments
    all have the same duration.

    Parameters
    ----------
    moduli : np.ndarray(real) or Tensor(real)
        The moduli :math:`\{\Omega_n\}` of the values of :math:`N` constant
        segments. These can represent either the moduli of a single
        sequence of segment values or of a batch of them. To provide a
        batch of sequences of segment values of shape
        :math:`B_1 \times \ldots \times B_n`, represent these moduli as a
        tensor of shape :math:`B_1 \times \ldots \times B_n \times N`.
    phases : np.ndarray(real) or Tensor(real)
        The phases :math:`\{\phi_n\}` of the complex segment values. Must
        have the same length as `moduli` (or same shape, if you're
        providing a batch).
    duration : float
        The total duration :math:`\tau` of the signal.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc(complex)
        The piecewise-constant function of time :math:`v(t)`, satisfying
        :math:`v(t)=\Omega_n e^{i\phi_n}` for :math:`t_{n-1}\leq t\leq t_n`,
        where :math:`t_n=n\tau/N` (where :math:`N` is the number of
        values in :math:`\{\Omega_n\}` and :math:`\{\phi_n\}`). If you
        provide a batch of `moduli` and `phases`, the returned `Pwc`
        represents a corresponding batch of
        :math:`B_1 \times \ldots \times B_n` functions :math:`v(t)`.

    See Also
    --------
    Graph.pwc_signal : Create `Pwc` signals from (possibly complex) values.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a complex piecewise-constant signal with batched moduli and phases.

    >>> moduli = np.array([[1, 2], [3, 4]])
    >>> phases = np.array([[0.1, 0.2], [0.5, 0.7]])
    >>> graph.complex_pwc_signal(moduli=moduli, phases=phases, duration=0.2, name="signal")
    <Pwc: name="signal", operation_name="complex_pwc_signal", value_shape=(), batch_shape=(2,)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="signal")
    >>> result["output"]["signal"]
    {
        'durations': array([0.1, 0.1]),
        'values': array([
            [0.99500417+0.09983342j, 1.96013316+0.39733866j],
            [2.63274769+1.43827662j, 3.05936875+2.57687075j]
        ]),
        'time_dimension': 1
    }

    See more examples in the `Design robust single-qubit gates using computational graphs
    <https://docs.q-ctrl.com/boulder-opal/tutorials/design-robust-single-qubit-gates-
    using-computational-graphs>`_ tutorial.
    """

    name = "complex_pwc_signal"
    args = [
        forge.arg("moduli", type=TensorLike),
        forge.arg("phases", type=TensorLike),
        forge.arg("duration", type=float),
    ]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        moduli = kwargs.get("moduli")
        phases = kwargs.get("phases")
        duration = kwargs.get("duration")
        check_numeric_numpy_array(moduli, "moduli")
        check_numeric_numpy_array(phases, "phases")
        check_duration(duration, "duration")
        moduli_shape = validate_shape(moduli, "moduli")
        phases_shape = validate_shape(phases, "phases")
        check_argument(
            len(moduli_shape) > 0,
            "The shape of moduli must have at least one dimension.",
            {"moduli": moduli},
        )
        check_argument(
            moduli_shape == phases_shape,
            "The shape of moduli and phases must be equal.",
            {"moduli": moduli, "phases": phases},
        )
        durations = duration / moduli_shape[-1] * np.ones(moduli_shape[-1])
        return Pwc(
            _operation,
            value_shape=(),
            durations=durations,
            batch_shape=moduli_shape[:-1],
        )


class PwcOperator(Node):
    r"""
    Create a constant operator multiplied by a piecewise-constant signal.

    Parameters
    ----------
    signal : Pwc
        The piecewise-constant signal :math:`a(t)`, or a batch of
        piecewise-constant signals.
    operator : np.ndarray or Tensor
        The operator :math:`A`. It must have two equal dimensions.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant operator :math:`a(t)A` (or a batch of
        piecewise-constant operators, if you provide a batch of
        piecewise-constant signals).

    See Also
    --------
    Graph.complex_pwc_signal : Create complex `Pwc` signals from their moduli and phases.
    Graph.constant_pwc_operator : Create constant `Pwc`\s.
    Graph.hermitian_part : Hermitian part of an operator.
    Graph.pwc : Create piecewise-constant functions.
    Graph.pwc_signal : Create `Pwc` signals from (possibly complex) values.
    Graph.pwc_sum : Sum multiple `Pwc`\s.
    Graph.stf_operator : Corresponding operation for `Stf`\s.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a piecewise-constant operator with non-uniform segment durations.

    >>> sigma_z = np.array([[1.0, 0.0],[0.0, -1.0]])
    >>> graph.pwc_operator(
    ...     signal=graph.pwc(durations=np.array([0.1, 0.2]), values=np.array([1, 2])),
    ...     operator=sigma_z,
    ...     name="operator",
    ... )
    <Pwc: name="operator", operation_name="pwc_operator", value_shape=(2, 2), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="operator")
    >>> result["output"]["operator"]
    {
        'durations': array([0.1, 0.2]),
        'values': array([
            [[ 1.,  0.], [ 0., -1.]],
            [[ 2.,  0.], [ 0., -2.]]
        ]),
        'time_dimension': 0
    }

    See more examples in the `How to represent quantum systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-represent-quantum-
    systems-using-graphs>`_ user guide.
    """

    name = "pwc_operator"
    args = [forge.arg("signal", type=Pwc), forge.arg("operator", type=TensorLike)]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        signal = kwargs.get("signal")
        operator = kwargs.get("operator")
        check_argument(
            isinstance(signal, Pwc), "The signal must be a Pwc.", {"signal": signal}
        )
        batch_shape, signal_value_shape = validate_batch_and_value_shapes(
            signal, "signal"
        )
        check_argument(
            not signal_value_shape,
            "The signal must be scalar-valued.",
            {"signal": signal},
        )

        value_shape = validate_shape(operator, "operator")
        check_operator(operator, "operator")
        check_argument(
            len(value_shape) == 2,
            "The operator must be a matrix, not a batch.",
            {"operator": operator},
            extras={"operator.shape": value_shape},
        )
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=signal.durations,
            batch_shape=batch_shape,
        )


class ConstantPwcOperator(Node):
    r"""
    Create a constant piecewise-constant operator over a specified duration.

    Parameters
    ----------
    duration : float
        The duration :math:`\tau` for the resulting piecewise-constant
        operator.
    operator : np.ndarray or Tensor
        The operator :math:`A`, or a batch of operators. It must have at
        least two dimensions, and its last two dimensions must be equal.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The constant operator :math:`t\mapsto A` (for :math:`0\leq t\leq\tau`)
        (or a batch of constant operators, if you provide a batch of operators).

    See Also
    --------
    Graph.constant_stf_operator : Corresponding operation for `Stf`\s.
    Graph.hermitian_part : Hermitian part of an operator.
    Graph.pwc_operator : Create `Pwc` operators.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a Hamiltonian from a batched constant operator.

    >>> sigma_x = np.array([[0.0, 1.0], [1.0, 0.0]])
    >>> sigma_z = np.array([[1.0, 0.0], [0.0, -1.0]])
    >>> batched_operators = np.asarray([sigma_x, sigma_z])
    >>> graph.constant_pwc_operator(duration=0.1, operator=batched_operators,  name="op")
    <Pwc: name="op", operation_name="constant_pwc_operator", value_shape=(2, 2), batch_shape=(2,)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="op")
    >>> result["output"]["op"]
    {
        'durations': array([0.1]),
        'values': array([
            [[[ 0.,  1.], [ 1.,  0.]]],
            [[[ 1.,  0.], [ 0., -1.]]]
        ]),
        'time_dimension': 1
    }

    See more examples in the `How to represent quantum systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-represent-quantum-
    systems-using-graphs>`_ user guide.
    """

    name = "constant_pwc_operator"
    args = [forge.arg("duration", type=float), forge.arg("operator", type=TensorLike)]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        duration = kwargs.get("duration")
        operator = kwargs.get("operator")
        check_duration(duration, "duration")
        shape = validate_shape(operator, "operator")
        check_operator(operator, "operator")
        return Pwc(
            _operation,
            value_shape=shape[-2:],
            durations=np.array([duration]),
            batch_shape=shape[:-2],
        )


class ConstantPwc(Node):
    r"""
    Create a piecewise-constant function of time that is constant over a specified duration.

    Parameters
    ----------
    constant : number or np.ndarray or Tensor
        The value :math:`c` of the function on the constant segment.
        To create a batch of :math:`B_1 \times \ldots \times B_n` piecewise-constant
        functions of shape :math:`D_1 \times \ldots \times D_m`, provide this `constant`
        parameter as an object of shape
        :math:`B_1\times\ldots\times B_n\times D_1\times\ldots\times D_m`.
    duration : float
        The duration :math:`\tau` for the resulting piecewise-constant function.
    batch_dimension_count : int, optional
        The number of batch dimensions, :math:`n` in `constant`.
        If provided, the first :math:`n` dimensions of `constant` are considered batch dimensions.
        Defaults to 0, which corresponds to no batch.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The constant function :math:`f(t) = c` (for :math:`0\leq t\leq\tau`)
        (or a batch of constant functions, if you provide `batch_dimension_count`).

    See Also
    --------
    Graph.constant_pwc_operator : Create constant `Pwc` operators.
    Graph.constant_stf : Corresponding operation for `Stf`\s.
    Graph.pwc : Create piecewise-constant functions.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a batched piecewise-constant function.

    >>> constant = np.arange(12).reshape((2, 2, 3))
    >>> graph.constant_pwc(
    ...     constant=constant, duration=0.1, batch_dimension_count=1, name="constant"
    ... )
    <Pwc: name="constant", operation_name="constant_pwc", value_shape=(2, 3), batch_shape=(2,)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="constant")
    >>> result["output"]["constant"]
    {
        'durations': array([0.1]),
        'values': array([
            [[[ 0.,  1.,  2.], [ 3.,  4.,  5.]]],
            [[[ 6.,  7.,  8.], [ 9., 10., 11.]]]
        ]),
        'time_dimension': 1
    }

    See more examples in the `Simulate the dynamics of a single qubit using computational graphs
    <https://docs.q-ctrl.com/boulder-opal/tutorials/simulate-the-dynamics-of-a-single-qubit-
    using-computational-graphs>`_ tutorial.
    """

    name = "constant_pwc"
    args = [
        forge.arg("constant", type=Union[float, complex, np.ndarray, Tensor]),
        forge.arg("duration", type=float),
        forge.arg("batch_dimension_count", type=int, default=0),
    ]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        constant = kwargs.get("constant")
        duration = kwargs.get("duration")
        batch_dimension_count = kwargs.get("batch_dimension_count")

        check_numeric_numpy_array(constant, "constant")
        check_duration(duration, "duration")
        check_argument_integer(batch_dimension_count, "batch_dimension_count")

        shape = validate_shape(constant, "constant")
        check_argument(
            len(shape) >= batch_dimension_count,
            "The number of batch dimensions must not be larger than the number "
            "of dimensions of the input constant.",
            {"constant": constant, "batch_dimension_count": batch_dimension_count},
            extras={"Number of value dimensions": len(shape)},
        )
        check_argument(
            batch_dimension_count >= 0,
            "The number of batch dimensions must not be negative.",
            {"batch_dimension_count": batch_dimension_count},
        )

        return Pwc(
            _operation,
            value_shape=shape[batch_dimension_count:],
            durations=np.array([duration]),
            batch_shape=shape[:batch_dimension_count],
        )


class PwcSum(Node):
    r"""
    Create the sum of multiple piecewise-constant terms.

    Parameters
    ----------
    terms : list[Pwc]
        The individual piecewise-constant terms :math:`\{v_j(t)\}` to sum. All
        terms must have the same duration, values of the same shape, and the
        same batch shape, but may have different segmentations (different
        numbers of segments of different durations) and different dtypes of
        segment values.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function (or batch of functions) of time
        :math:`\sum_j v_j(t)`. Its values have the same shape as the values of
        each of the `terms` that you provided. If each of the `terms` represents
        a batch of functions, this result represents a batch of functions with
        the same batch shape. If any `term` has complex-valued segments, the
        value of the returned Pwc is complex, otherwise is float.

    See Also
    --------
    Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
    Graph.pwc : Create piecewise-constant functions.
    Graph.pwc_operator : Create `Pwc` operators.
    Graph.pwc_signal : Create `Pwc` signals from (possibly complex) values.
    Graph.stf_sum : Corresponding operation for `Stf`\s.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Sum a list of piecewise-constant terms of different durations.

    >>> x = graph.pwc(durations=np.array([0.1, 0.3]), values=np.array([1, 2]))
    >>> y = graph.pwc(durations=np.array([0.2, 0.2]), values=np.array([3, 1]))
    >>> graph.pwc_sum([x, y], name="sum")
    <Pwc: name="sum", operation_name="pwc_sum", value_shape=(), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="sum")
    >>> result["output"]["sum"]
    {
        'durations': array([0.1, 0.1, 0.2]),
        'values': array([4., 5., 3.]),
        'time_dimension': 0
    }

    See more examples in the `How to optimize controls robust to strong noise sources
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-robust-to-
    strong-noise-sources>`_ user guide.
    """

    name = "pwc_sum"
    args = [forge.arg("terms", type=List[Pwc])]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        terms = kwargs.get("terms")
        check_argument_iterable(terms, "terms")
        check_argument(
            len(terms) > 0, "You must provide at least one term.", {"terms": terms}
        )
        check_argument(
            all(isinstance(term, Pwc) for term in terms),
            "Each of the terms must be a Pwc.",
            {"terms": terms},
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(terms[0], "terms[0]")
        check_argument(
            all(term.value_shape == value_shape for term in terms[1:]),
            "All the terms must have the same value shape.",
            {"terms": terms},
            extras={"value shapes of terms": [term.value_shape for term in terms]},
        )
        check_argument(
            all(term.batch_shape == batch_shape for term in terms[1:]),
            "All the terms must have the same batch shape.",
            {"terms": terms},
            extras={"batch shapes of terms": [term.batch_shape for term in terms]},
        )
        check_argument(
            all(
                np.allclose(np.sum(terms[0].durations), np.sum(term.durations))
                for term in terms
            ),
            "All the terms must have the same total duration.",
            {"terms": terms},
            extras={
                "total duration of each term": [
                    np.sum(term.durations) for term in terms
                ]
            },
        )
        durations = mesh_pwc_durations(terms)

        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )


class TimeReversePwc(Node):
    r"""
    Reverse in time a piecewise-constant function.

    Parameters
    ----------
    pwc : Pwc
        The piecewise-constant function :math:`v(t)` to reverse.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function :math:`w(t)` defined by
        :math:`w(t)=v(\tau-t)` for :math:`0\leq t\leq \tau`, where
        :math:`\tau` is the duration of :math:`v(t)`.

    See Also
    --------
    Graph.symmetrize_pwc : Symmetrize `Pwc`\s.
    Graph.time_concatenate_pwc : Concatenate `Pwc`\s in time.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Reverse a piecewise constant function.

    >>> x = graph.pwc(durations=np.array([0.1, 0.5, 0.3]), values=np.array([1, 2, 3]))
    >>> graph.time_reverse_pwc(x, name="reverse")
    <Pwc: name="reverse", operation_name="time_reverse_pwc", value_shape=(), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="reverse")
    >>> result["output"]["reverse"]
    {
        'durations': array([0.3, 0.5, 0.1]),
        'values': array([3., 2., 1.]),
        'time_dimension': 0
    }
    """

    name = "time_reverse_pwc"
    args = [forge.arg("pwc", type=Pwc)]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        pwc = kwargs.get("pwc")
        check_argument(
            isinstance(pwc, Pwc), "The pwc parameter must be a Pwc.", {"pwc": pwc}
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(pwc, "pwc")
        reversed_durations = pwc.durations[::-1]
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=reversed_durations,
            batch_shape=batch_shape,
        )


class SymmetrizePwc(Node):
    r"""
    Create the symmetrization of a piecewise-constant function.

    Parameters
    ----------
    pwc : Pwc
        The piecewise-constant function :math:`v(t)` to symmetrize.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function :math:`w(t)` defined by
        :math:`w(t)=v(t)` for :math:`0\leq t\leq \tau` and
        :math:`w(t)=v(2\tau-t)` for :math:`\tau\leq t\leq 2\tau`, where
        :math:`\tau` is the duration of :math:`v(t)`.

    See Also
    --------
    Graph.pwc_signal : Create `Pwc` signals from (possibly complex) values.
    Graph.time_reverse_pwc : Reverse `Pwc`\s in time.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create a symmetric piecewise-constant function.

    >>> x = graph.pwc(durations=np.array([0.1, 0.3]), values=np.array([1, 2]))
    >>> graph.symmetrize_pwc(x, name="symmetrize")
    <Pwc: name="symmetrize", operation_name="symmetrize_pwc", value_shape=(), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="symmetrize")
    >>> result["output"]["symmetrize"]
    {
        'durations': array([0.1, 0.3, 0.3, 0.1]),
        'values': array([1., 2., 2., 1.]),
        'time_dimension': 0
    }

    See more examples in the `How to optimize controls with time symmetrization
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-with-
    time-symmetrization>`_ user guide.
    """

    name = "symmetrize_pwc"
    args = [forge.arg("pwc", type=Pwc)]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        pwc = kwargs.get("pwc")
        check_argument(
            isinstance(pwc, Pwc), "The pwc parameter must be a Pwc.", {"pwc": pwc}
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(pwc, "pwc")
        symmetrized_durations = np.concatenate((pwc.durations, pwc.durations[::-1]))
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=symmetrized_durations,
            batch_shape=batch_shape,
        )


class TimeConcatenatePwc(Node):
    r"""
    Concatenate multiple piecewise-constant functions in the time dimension.

    Parameters
    ----------
    pwc_list : list[Pwc]
        The individual piecewise-constant functions :math:`\{A_i(t)\}` to concatenate.
        All the functions must have the same value shape, and can have broadcastable batch shapes.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The concatenated piecewise-constant function (or batch of functions).

    See Also
    --------
    Graph.pwc : Create piecewise-constant functions.
    Graph.pwc_sum : Sum multiple `Pwc`\s.
    Graph.symmetrize_pwc : Symmetrize `Pwc`\s.
    Graph.time_reverse_pwc : Reverse `Pwc`\s in time.

    Notes
    -----
    The function resulting from the concatenation is

    .. math::
        C(t) = \begin{cases}
        A_0(t) & \mathrm{for} & 0 < t < \tau_0
        \\
        A_1(t - \tau_0) & \mathrm{for} & \tau_0 < t < \tau_0 + \tau_1
        \\
        A_2(t - \tau_0 - \tau_1) & \mathrm{for} & \tau_0 + \tau_1 < t < \tau_0 + \tau_1 + \tau_2
        \\
        & \vdots &
        \end{cases}

    where :math:`\tau_i` is the duration of the i-th function.

    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Concatenate two piecewise-constant functions.

    >>> pwc1 = graph.pwc(durations=np.array([0.2, 0.5]), values=np.array([1, 2]))
    >>> pwc2 = graph.pwc(durations=np.array([0.7, 0.9]), values=np.array([3, 4]))
    >>> graph.time_concatenate_pwc([pwc1, pwc2], name="concat")
    <Pwc: name="concat", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="concat")
    >>> result["output"]["concat"]
    {
        'durations': array([0.2, 0.5, 0.7, 0.9]),
        'values': array([1., 2., 3., 4.]),
        'time_dimension': 0
    }
    """

    name = "time_concatenate_pwc"
    args = [forge.arg("pwc_list", type=List[Pwc])]
    rtype = Pwc
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        pwc_list = kwargs.get("pwc_list")
        check_argument_iterable(pwc_list, "pwc_list")
        check_argument(
            len(pwc_list) > 0,
            "You must provide at least one element.",
            {"pwc_list": pwc_list},
        )
        check_argument(
            all(isinstance(pwc, Pwc) for pwc in pwc_list),
            "Each of the elements to concatenate must be a Pwc.",
            {"pwc_list": pwc_list},
        )
        _, value_shape = validate_batch_and_value_shapes(pwc_list[0], "pwc_list[0]")
        check_argument(
            all((pwc.value_shape == value_shape) for pwc in pwc_list[1:]),
            "All the Pwcs must have the same value shape.",
            {"pwc_list": pwc_list},
        )
        batch_shape = get_broadcasted_shape(*[pwc.batch_shape for pwc in pwc_list])
        check_argument(
            batch_shape is not None,
            "All the Pwcs must have broadcastable batch shapes.",
            {"pwc_list": pwc_list},
        )

        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=np.concatenate([pwc.durations for pwc in pwc_list]),
            batch_shape=batch_shape,
        )


class TimeEvolutionOperatorsPwc(Node):
    """
    Calculate the unitary time-evolution operators for a system defined by a piecewise-constant
    Hamiltonian.

    Parameters
    ----------
    hamiltonian : Pwc
        The control Hamiltonian, or batch of control Hamiltonians.
    sample_times : list or tuple or np.ndarray(1D, real)
        The N times at which you want to sample the unitaries. Must be ordered and contain
        at least one element.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        Tensor of shape ``(..., N, D, D)``, representing the unitary time evolution.
        The n-th element (along the -3 dimension) represents the unitary (or batch of unitaries)
        from t = 0 to ``sample_times[n]``.

    See Also
    --------
    Graph.state_evolution_pwc : Evolve a quantum state.
    Graph.time_evolution_operators_stf : Corresponding operation for `Stf` Hamiltonians.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Simulate the dynamics of a single qubit, where a constant drive rotates the
    qubit along the x-axis.

    >>> initial_state = np.array([1, 0])
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> duration = np.pi
    >>> hamiltonian = graph.constant_pwc_operator(duration=duration, operator=sigma_x / 2)
    >>> graph.time_evolution_operators_pwc(
    ...     hamiltonian=hamiltonian, sample_times=[duration], name="unitaries"
    ... )
    <Tensor: name="unitaries", operation_name="time_evolution_operators_pwc", shape=(1, 2, 2)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="unitaries")
    >>> result["output"]["unitaries"]["value"].dot(initial_state)
    array([[5.0532155e-16+0.j, 0.0000000e+00-1.j]])

    See more examples in the `How to simulate quantum dynamics for noiseless systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-quantum-dynamics-for-
    noiseless-systems-using-graphs>`_ user guide.
    """

    name = "time_evolution_operators_pwc"
    args = [
        forge.arg("hamiltonian", type=Pwc),
        forge.arg("sample_times", type=Union[list, tuple, np.ndarray]),
    ]
    rtype = Tensor
    categories = [Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        hamiltonian = kwargs.get("hamiltonian")
        sample_times = kwargs.get("sample_times")
        check_argument(
            isinstance(hamiltonian, Pwc),
            "The Hamiltonian must be a Pwc.",
            {"hamiltonian": hamiltonian},
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(
            hamiltonian, "hamiltonian"
        )
        sample_times = np.asarray(sample_times)
        check_sample_times_with_bounds(
            sample_times, "sample_times", hamiltonian, "hamiltonian"
        )
        validate_hamiltonian(hamiltonian, "hamiltonian")
        time_count = len(sample_times)
        shape = batch_shape + (time_count,) + value_shape
        return Tensor(_operation, shape=shape)


class SamplePwc(Node):
    r"""
    Sample a Pwc at the given times.

    Parameters
    ----------
    pwc : Pwc
        The Pwc to sample.
    sample_times : list or tuple or np.ndarray(1D, real)
        The times at which you want to sample the Pwc. Must be ordered, contain
        at least one element, and lie between 0 and the duration of the Pwc.
        For a sample time :math:`t` the returned value
        lies in the half open interval :math:`t_{j-1} < t \leq t_j`, where
        :math:`t_j` indicates the boundary of a segment.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The values of the Pwc at the given times.

    See Also
    --------
    Graph.sample_stf : Sample an `Stf` at given times.

    Notes
    -----
    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "sample_pwc"
    args = [
        forge.arg("pwc", type=Pwc),
        forge.arg("sample_times", type=Union[list, tuple, np.ndarray]),
    ]
    rtype = Tensor
    categories = [Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        pwc = kwargs.get("pwc")
        check_argument(isinstance(pwc, Pwc), "The pwc must be a Pwc.", {"pwc": pwc})
        batch_shape, value_shape = validate_batch_and_value_shapes(pwc, "pwc")
        sample_times = kwargs.get("sample_times")
        sample_times = np.asarray(sample_times)
        check_sample_times_with_bounds(sample_times, "sample_times", pwc, "pwc")
        check_argument(
            sample_times[0] >= 0,
            "The first sample time can't be smaller than zero.",
            {"sample_times": sample_times},
        )
        check_argument(
            sample_times[-1] <= np.sum(pwc.durations),
            "The last sample time can't be larger than the duration of the Pwc.",
            {"sample_times": sample_times},
            extras={"Pwc durations": pwc.durations},
        )
        time_count = len(sample_times)
        shape = batch_shape + (time_count,) + value_shape
        return Tensor(_operation, shape=shape)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_hermitian,
    check_argument_integer,
    check_duration,
    check_numeric_numpy_array,
    check_operator,
)
from scipy.sparse import spmatrix

from .documentation import Category
from .node_data import (
    Pwc,
    SparsePwc,
    Tensor,
)
from .types import TensorLike
from .utils import (
    check_sample_times_with_bounds,
    mesh_pwc_durations,
    validate_hamiltonian,
    validate_shape,
)


class SparsePwcOperator(Node):
    r"""
    Create a sparse piecewise-constant operator (sparse-matrix-valued function of time).

    Each of the piecewise-constant segments (time periods) is a scalar multiple
    of the operator.

    Parameters
    ----------
    signal : Pwc
        The scalar-valued piecewise-constant function of time :math:`a(t)`.
    operator : numpy.ndarray or scipy.sparse.spmatrix or Tensor
        The sparse operator :math:`A` to be scaled over time. If you pass a Tensor or NumPy array,
        it will be internally converted into a sparse representation.

    Returns
    -------
    SparsePwc
        The piecewise-constant sparse operator :math:`a(t)A`.

    See Also
    --------
    Graph.constant_sparse_pwc_operator : Create constant `SparsePwc`\s.
    Graph.density_matrix_evolution_pwc : Evolve a quantum state in an open system.
    Graph.pwc_operator : Corresponding operation for `Pwc`\s.
    Graph.sparse_pwc_hermitian_part : Hermitian part of a `SparsePwc` operator.
    Graph.sparse_pwc_sum : Sum multiple `SparsePwc`\s.
    Graph.state_evolution_pwc : Evolve a quantum state.

    Examples
    --------
    Create a sparse PWC operator.

    >>> from scipy.sparse import coo_matrix
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> signal = graph.pwc_signal(values=np.array([1, 2, 3]), duration=0.1)
    >>> graph.sparse_pwc_operator(signal=signal, operator=coo_matrix(sigma_x))
    <SparsePwc: operation_name="sparse_pwc_operator", value_shape=(2, 2)>

    See more examples in the `How to simulate large open system dynamics
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
    user guide.
    """

    name = "sparse_pwc_operator"
    args = [
        forge.arg("signal", type=Pwc),
        forge.arg("operator", type=Union[np.ndarray, spmatrix, Tensor]),
    ]
    kwargs = {}  # SparsePwc doesn't accept `name` as an argument.
    rtype = SparsePwc
    categories = [Category.LARGE_SYSTEMS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        signal = kwargs.get("signal")
        operator = kwargs.get("operator")
        check_argument(
            signal.value_shape == (),
            "The signal must be scalar-valued.",
            {"signal": signal},
            extras={"signal.value_shape": signal.value_shape},
        )
        batch_shape = signal.batch_shape
        check_argument(
            batch_shape == (),
            "Batching is not supported for SparsePwc nodes.",
            {"signal": signal},
            extras={"signal.batch_shape": batch_shape},
        )
        check_argument(
            isinstance(operator, (np.ndarray, spmatrix, Tensor)),
            "The operator must be a NumPy array, or a SciPy sparse matrix, or a Tensor.",
            {"operator": operator},
        )
        operator_shape = validate_shape(operator, "operator")
        check_argument(
            len(operator_shape) == 2 and operator_shape[0] == operator_shape[1],
            "Operator must be a 2D square matrix.",
            {"operator": operator},
            extras={"operator.shape": operator_shape},
        )
        return SparsePwc(
            _operation, value_shape=operator_shape, durations=signal.durations
        )


class ConstantSparsePwcOperator(Node):
    r"""
    Create a constant sparse piecewise-constant operator over a specified duration.

    Parameters
    ----------
    duration : float
        The duration :math:`\tau` for the resulting piecewise-constant operator.
    operator : numpy.ndarray or scipy.sparse.spmatrix or Tensor
        The sparse operator :math:`A`. If you pass a Tensor or NumPy array,
        it will be internally converted into a sparse representation.

    Returns
    -------
    SparsePwc
        The constant operator :math:`t\mapsto A` (for :math:`0\leq t\leq\tau`).

    See Also
    --------
    Graph.constant_pwc_operator : Corresponding operation for `Pwc`\s.
    Graph.sparse_pwc_hermitian_part : Hermitian part of a `SparsePwc` operator.
    Graph.sparse_pwc_operator : Create `SparsePwc` operators.
    Graph.sparse_pwc_sum : Sum multiple `SparsePwc`\s.

    Examples
    --------
    Create a constant sparse PWC operator.

    >>> from scipy.sparse import coo_matrix
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_x))
    <SparsePwc: operation_name="constant_sparse_pwc_operator", value_shape=(2, 2)>

    See more examples in the `How to simulate large open system dynamics
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
    user guide.
    """

    name = "constant_sparse_pwc_operator"
    args = [
        forge.arg("duration", type=float),
        forge.arg("operator", type=Union[np.ndarray, spmatrix, Tensor]),
    ]
    kwargs = {}  # SparsePwc doesn't accept `name` as an argument.
    rtype = SparsePwc
    categories = [Category.LARGE_SYSTEMS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        duration = kwargs.get("duration")
        operator = kwargs.get("operator")
        check_duration(duration, "duration")
        check_argument(
            isinstance(operator, (np.ndarray, spmatrix, Tensor)),
            "The operator must be a NumPy array, a SciPy sparse matrix, or a Tensor.",
            {"operator": operator},
        )
        operator_shape = validate_shape(operator, "operator")
        check_argument(
            len(operator_shape) == 2 and operator_shape[0] == operator_shape[1],
            "Operator must be a 2D square matrix.",
            {"operator": operator},
            extras={"operator.shape": operator_shape},
        )
        return SparsePwc(
            _operation, value_shape=operator_shape, durations=np.array([duration])
        )


class SparsePwcSum(Node):
    r"""
    Create the sum of multiple sparse-matrix-valued piecewise-constant functions.

    Parameters
    ----------
    terms : list[SparsePwc]
        The individual piecewise-constant terms :math:`\{v_j(t)\}` to sum.
        All terms must be sparse, have values of the same shape,
        and have the same total duration
        but may have different numbers of segments of different durations.

    Returns
    -------
    SparsePwc
        The piecewise-constant function of time :math:`\sum_j v_j(t)`. It
        has the same shape as each of the `terms` that you provided.

    See Also
    --------
    Graph.constant_sparse_pwc_operator : Create constant `SparsePwc`\s.
    Graph.pwc_sum : Corresponding operation for `Pwc`\s.
    Graph.sparse_pwc_operator : Create `SparsePwc` operators.

    Examples
    --------
    Sum two sparse PWC operators.

    >>> from scipy.sparse import coo_matrix
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> sigma_y = np.array([[0, -1j], [1j, 0]])
    >>> sp_x = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_x))
    >>> sp_y = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_y))
    >>> graph.sparse_pwc_sum([sp_x, sp_y])
    <SparsePwc: operation_name="sparse_pwc_sum", value_shape=(2, 2)>

    See more examples in the `How to simulate large open system dynamics
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
    user guide.
    """

    name = "sparse_pwc_sum"
    args = [forge.arg("terms", type=List[SparsePwc])]
    kwargs = {}  # SparsePwcSum doesn't accept `name` as an argument.
    rtype = SparsePwc
    categories = [Category.LARGE_SYSTEMS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        terms = kwargs.get("terms")

        check_argument(
            isinstance(terms, list),
            "The terms must be provided as a list.",
            {"terms": terms},
            extras={"type(terms)": type(terms)},
        )
        check_argument(
            len(terms) > 0, "At least one term must be provided.", {"terms": terms}
        )
        shape_0 = terms[0].value_shape
        for i, term in enumerate(terms):
            check_argument(
                isinstance(term, SparsePwc),
                "All the terms must be SparsePwc.",
                {"terms": terms},
                extras={f"terms[{i}]": term},
            )
            shape = term.value_shape
            check_argument(
                shape == shape_0,
                "All the terms must have the same shape.",
                {"terms": terms},
                extras={
                    f"terms[{0}].value_shape": shape_0,
                    f"terms[{i}].value_shape": shape,
                },
            )
        durations = mesh_pwc_durations(terms)
        return SparsePwc(_operation, value_shape=shape_0, durations=durations)


class SparsePwcHermitianPart(Node):
    r"""
    Create the Hermitian part of a piecewise-constant operator.

    Parameters
    ----------
    operator : SparsePwc
        The operator :math:`A(t)`.

    Returns
    -------
    SparsePwc
        The Hermitian part :math:`\frac{1}{2}(A(t)+A^\dagger(t))`.

    See Also
    --------
    Graph.hermitian_part : Hermitian part of an operator.
    Graph.sparse_pwc_operator : Create `SparsePwc`\s.

    Examples
    --------
    Create a Hermitian sparse PWC operator.

    >>> from scipy.sparse import coo_matrix
    >>> sigma_m = np.array([[0, 1], [0, 0]])
    >>> sp_m = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_m))
    >>> graph.sparse_pwc_hermitian_part(sp_m)
    <SparsePwc: operation_name="sparse_pwc_hermitian_part", value_shape=(2, 2)>
    """

    name = "sparse_pwc_hermitian_part"
    args = [forge.arg("operator", type=SparsePwc)]
    kwargs = {}  # SparsePwc is not fetchable.
    rtype = SparsePwc
    categories = [Category.LARGE_SYSTEMS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        operator = kwargs.get("operator")
        check_argument(
            isinstance(operator, SparsePwc),
            "The operator must be a SparsePwc.",
            {"operator": operator},
        )
        return SparsePwc(
            _operation, value_shape=operator.value_shape, durations=operator.durations
        )


class StateEvolutionPwc(Node):
    r"""
    Calculate the time evolution of a state generated by a piecewise-constant
    Hamiltonian by using the Lanczos method.

    Parameters
    ----------
    initial_state : Tensor or np.ndarray
        The initial state as a Tensor or np.ndarray of shape ``(D,)``.
    hamiltonian : Pwc or SparsePwc
        The control Hamiltonian. Uses sparse matrix multiplication if of type
        `SparsePwc`, which can be more efficient for large operators that are
        relatively sparse (contain mostly zeros).
    krylov_subspace_dimension : Tensor or int
        The dimension of the Krylov subspace `k` for the Lanczos method.
    sample_times : list or tuple or np.ndarray(1D, real) or None, optional
        The N times at which you want to sample the state. Elements must be non-negative
        and strictly increasing, with a supremum that is the duration of the `hamiltonian`.
        If omitted only the evolved state at the final time of the control Hamiltonian
        is returned.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        Tensor of shape ``(N, D)`` or ``(D,)`` if `sample_times` is omitted
        representing the state evolution. The n-th element (along the first
        dimension) represents the state at ``sample_times[n]`` evolved from
        the initial state.

    Warnings
    --------
    This calculation can be relatively inefficient for small systems (very roughly
    speaking, when the dimension of your Hilbert space is less than around 100; the
    exact cutoff depends on the specifics of your problem though). You should generally
    first try using `time_evolution_operators_pwc` to get the full time evolution
    operators (and evolve your state using those), and only switch to this method
    if that approach proves too slow or memory intensive. See the
    `How to simulate quantum dynamics for noiseless systems using graphs`_ user guide
    for an example of calculating state evolution with `time_evolution_operators_pwc`.

    .. _How to simulate quantum dynamics for noiseless systems using graphs:
        https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-quantum-
        dynamics-for-noiseless-systems-using-graphs

    See Also
    --------
    Graph.density_matrix_evolution_pwc : Corresponding operation for open systems.
    Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
    Graph.estimated_krylov_subspace_dimension_lanczos :
        Obtain a Krylov subspace dimension to use with this integrator.
    Graph.sparse_pwc_operator : Create `SparsePwc` operators.
    Graph.time_evolution_operators_pwc :
        Unitary time evolution operators for quantum systems with `Pwc` Hamiltonians.

    Notes
    -----
    The Lanczos algorithm calculates the unitary evolution of a state in the Krylov
    subspace. This subspace is spanned by the states resulting from applying the first
    `k` powers of the Hamiltonian on the input state, with `k` being the subspace dimension,
    much smaller that the full Hilbert space dimension. This allows for an efficient
    state propagation in high-dimensional systems compared to calculating the full
    unitary operator.

    Moreover, this function uses sparse matrix multiplication when the Hamiltonian is passed as a
    `SparsePwc`. This can lead to more efficient calculations when they involve large operators that
    are relatively sparse (contain mostly zeros). In this case, the initial state is still a densely
    represented array or tensor.

    Note that increasing the density of `sample_times` does not affect the accuracy of the
    integration. However, increasing the Krylov subspace dimension or subdividing the Hamiltonian
    into shorter piecewise-constant segments can reduce the integration error, at the expense of
    longer computation times.

    Examples
    --------
    See example in the `How to optimize controls on large sparse Hamiltonians
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on-
    large-sparse-hamiltonians>`_ user guide.
    """

    name = "state_evolution_pwc"
    args = [
        forge.arg("initial_state", type=TensorLike),
        forge.arg("hamiltonian", type=Union[Pwc, SparsePwc]),
        forge.arg("krylov_subspace_dimension", type=Union[int, Tensor]),
        forge.arg(
            "sample_times", type=Optional[Union[list, tuple, np.ndarray]], default=None
        ),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        initial_state = kwargs.get("initial_state")
        state_shape = validate_shape(initial_state, "initial_state")
        check_argument(
            len(state_shape) == 1,
            "The initial state must be 1D.",
            {"initial_state": initial_state},
            extras={"initial_state.shape": state_shape},
        )
        hamiltonian = kwargs.get("hamiltonian")
        sample_times = kwargs.get("sample_times")
        check_numeric_numpy_array(initial_state, "initial_state")
        if sample_times is not None:
            sample_times = np.asarray(sample_times)
            check_sample_times_with_bounds(
                sample_times, "sample_times", hamiltonian, "hamiltonian"
            )
        check_argument(
            isinstance(hamiltonian, (SparsePwc, Pwc)),
            "The Hamiltonian must be a SparsePwc or a Pwc.",
            {"hamiltonian": hamiltonian},
        )
        if isinstance(hamiltonian, Pwc):
            check_argument(
                hamiltonian.batch_shape == (),
                "Hamiltonian cannot contain a batch.",
                {"hamiltonian": hamiltonian},
                extras={"hamiltonian.batch_shape": hamiltonian.batch_shape},
            )
        hamiltonian_value_shape = hamiltonian.value_shape
        validate_hamiltonian(hamiltonian, "hamiltonian")
        check_argument(
            state_shape[0] == hamiltonian_value_shape[-1],
            "The Hilbert space dimensions of the state and the Hamiltonian must be equal.",
            {"initial_state": initial_state, "hamiltonian": hamiltonian},
            extras={
                "initial_state.shape": state_shape,
                "hamiltonian.value_shape": hamiltonian_value_shape,
            },
        )
        if sample_times is None:
            shape = state_shape
        else:
            shape = (len(sample_times),) + state_shape
        return Tensor(_operation, shape=shape)


class EstimatedKrylovSubspaceDimensionLanczos(Node):
    r"""
    Calculate an appropriate Krylov subspace dimension (:math:`k`) to use in the Lanczos
    integrator while keeping the total error in the evolution below a given error tolerance.

    Note that you can provide your own estimation of the Hamiltonian spectral range or use the
    `spectral_range` operation to perform that calculation.

    Parameters
    ----------
    spectral_range : Tensor or float
        Estimated order of magnitude of Hamiltonian spectral range (difference
        between largest and smallest eigenvalues).
    duration : float
        The total evolution time.
    maximum_segment_duration : float
        The maximum duration of the piecewise-constant Hamiltonian segments.
    error_tolerance : float, optional
        Tolerance for the error in the integration, defined as the Frobenius norm of
        the vectorial difference between the exact state and the estimated state.
        Defaults to 1e-6.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        Recommended value of :math:`k` to use in a Lanczos integration with a Hamiltonian with a
        similar spectral range to the one passed.

    See Also
    --------
    Graph.spectral_range : Range of the eigenvalues of a Hermitian operator.
    Graph.state_evolution_pwc : Evolve a quantum state.

    Notes
    -----
    To provide the recommended :math:`k` parameter, this function uses the bound in the error for
    the Lanczos algorithm [1]_ [2]_ as an estimate for the error. For a single time step this gives

    .. math::
        \mathrm{error} \leq 12 \exp \left( - \frac{(w\tau)^2}{16 k} \right)
        \left (\frac{e w \tau}{ 4 k}  \right )^{k}

    where :math:`\tau` is the time step and :math:`w` is the spectral range of the Hamiltonian.

    As this bound overestimates the error, the actual resulting errors with the recommended
    parameter are expected to be (a few orders of magnitude) smaller than the requested tolerance.

    References
    ----------
    .. [1] `N. Del Buono and L. Lopez,
           Lect. Notes Comput. Sci. 2658, 111 (2003).
           <https://doi.org/10.1007/3-540-44862-4_13>`_

    .. [2] `M. Hochbruck and C. Lubich,
           SIAM J. Numer. Anal. 34, 1911 (1997).
           <https://doi.org/10.1137/S0036142995280572>`_

    Examples
    --------
    >>> graph.estimated_krylov_subspace_dimension_lanczos(
    ...     spectral_range=30.0,
    ...     duration=5e-7,
    ...     maximum_segment_duration=2.5e-8,
    ...     error_tolerance=1e-5,
    ...     name="dim",
    ... )
    <Tensor: name="dim", operation_name="estimated_krylov_subspace_dimension_lanczos", shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="dim")
    >>> result["output"]["dim"]["value"]
    2

    See more examples in the `How to optimize controls on large sparse Hamiltonians
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on-
    large-sparse-hamiltonians>`_ user guide.
    """

    name = "estimated_krylov_subspace_dimension_lanczos"
    args = [
        forge.arg("spectral_range", type=Union[Tensor, float]),
        forge.arg("duration", type=float),
        forge.arg("maximum_segment_duration", type=float),
        forge.arg("error_tolerance", type=float, default=1e-6),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        duration = kwargs.get("duration")
        maximum_segment_duration = kwargs.get("maximum_segment_duration")
        error_tolerance = kwargs.get("error_tolerance")
        check_argument(
            maximum_segment_duration > 0,
            "The maximum segment duration must be positive.",
            {"maximum_segment_duration": maximum_segment_duration},
        )
        check_argument(
            duration > 0, "The duration must be positive.", {"duration": duration}
        )
        check_argument(
            maximum_segment_duration <= duration,
            "The maximum segment duration must be less than or equal to duration.",
            {
                "maximum_segment_duration": maximum_segment_duration,
                "duration": duration,
            },
        )
        check_argument(
            error_tolerance > 0,
            "The error tolerance must be positive.",
            {"error_tolerance": error_tolerance},
        )

        return Tensor(_operation, shape=())


class SpectralRange(Node):
    r"""
    Obtain the range of the eigenvalues of a Hermitian operator.

    This function provides an estimate of the difference between the
    highest and the lowest eigenvalues of the operator. You can adjust its
    precision by modifying its default parameters.

    Parameters
    ----------
    operator : np.ndarray or scipy.sparse.spmatrix or Tensor
        The Hermitian operator :math:`M` whose range of eigenvalues you
        want to determine.
    iteration_count : int, optional
        The number of iterations :math:`N` in the calculation. Defaults to
        3000. Choose a higher number to improve the precision, or a smaller
        number to make the estimation run faster.
    seed : int or None, optional
        The random seed that the function uses to choose the initial random
        vector :math:`\left| r \right\rangle`. Defaults to None, which
        means that the function uses a different seed in each run.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor (scalar, real)
        The difference between the largest and the smallest eigenvalues of
        the operator.

    Warnings
    --------
    This calculation can be expensive, so we recommend that you run it
    before the optimization, if possible. You can do this by using a
    representative or a worst-case operator.

    Notes
    -----
    This function repeatedly multiplies the operator :math:`M` with a
    random vector :math:`\left| r \right\rangle`. In terms of the operator's
    eigenvalues :math:`\{ v_i \}` and eigenvectors
    :math:`\{\left|v_i \right\rangle\}`, the result of :math:`N` matrix
    multiplications is:

    .. math::
        M^N \left|r\right\rangle = \sum_i v_i^N \left|v_i\right\rangle
        \left\langle v_i \right. \left| r \right\rangle.

    For large :math:`N`, the term corresponding to the eigenvalue with
    largest absolute value :math:`V` will dominate the sum, as long as
    :math:`\left|r\right\rangle` has a non-zero overlap with its
    eigenvector. The function then retrieves the eigenvalue :math:`V` via:

    .. math::
        V \approx \frac{\left\langle r \right| M^{2N+1} \left| r
        \right\rangle}{\left\| M^N \left| r \right\rangle \right\|^2}.

    The same procedure applied to the matrix :math:`M-V` allows the function
    to find the eigenvalue at the opposite end of the spectral range.

    Examples
    --------
    >>> operator = np.diag([10, 40])
    >>> graph.spectral_range(operator, name="spectral_range")
    <Tensor: name="spectral_range", operation_name="spectral_range", shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="spectral_range")
    >>> result["output"]["spectral_range"]["value"]
    30.0

    See more examples in the `How to optimize controls on large sparse Hamiltonians
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on
    -large-sparse-hamiltonians>`_ user guide.
    """

    name = "spectral_range"
    args = [
        forge.arg("operator", type=Union[Tensor, np.ndarray, spmatrix]),
        forge.arg("iteration_count", type=int, default=3000),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    rtype = Tensor
    categories = [Category.LARGE_SYSTEMS, Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        operator = kwargs.get("operator")
        iteration_count = kwargs.get("iteration_count")
        check_operator(operator, "operator")
        operator_shape = validate_shape(operator, "operator")
        check_argument(
            len(operator_shape) == 2,
            "Batches of operators are not supported.",
            {"operator": operator},
            extras={"operator.shape": operator_shape},
        )
        check_argument_integer(iteration_count, "iteration_count")
        check_argument(
            iteration_count > 0,
            "The number of iterations must be greater than zero.",
            {"iteration_count": iteration_count},
        )
        if not isinstance(operator, Tensor):
            check_argument_hermitian(operator, "operator")
        return Tensor(_operation, shape=())
